"""
Basic Equations for solving shallow water problems
#####################
"""

from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.integrator import Integrator
from compyle.api import declare
from pysph.sph.wc.linalg import gj_solve, augmented_matrix

from numpy import sqrt, cos, sin, zeros, pi, exp
import numpy as np
import numpy
M_PI = pi


class CheckForParticlesToSplit(Equation):
    r"""Particles are tagged for splitting if the following condition is
    satisfied:

    .. math::
        (A_i > A_max) and (h_i < h_max) and (x_min < x_i < x_max) and (y_min <
        y_i < y_max)

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def __init__(self, dest, h_max=1e9, A_max=1e9, x_min=-1e9, x_max=1e9,
                 y_min=-1e9, y_max=1e9):
        r"""
        Parameters
        ----------
        h_max : float
            maximum smoothing length beyond which splitting is deactivated
        A_max : float
            maximum area beyond which splitting is activated
        x_min : float
            minimum distance along x-direction beyond which splitting is
            activated
        x_max : float
            maximum distance along x-direction beyond which splitting is
            deactivated
        y_min : float
            minimum distance along y-direction beyond which splitting is
            activated
        y_max : float
            maximum distance along y-direction beyond which splitting is
            deactivated

        """
        self.A_max = A_max
        self.h_max = h_max
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        super(CheckForParticlesToSplit, self).__init__(dest, None)

    def initialize(self, d_idx, d_A, d_h, d_x, d_y, d_pa_to_split):
        if (d_A[d_idx] > self.A_max and d_h[d_idx] < self.h_max
           and (self.x_min < d_x[d_idx] < self.x_max)
           and (self.y_min < d_y[d_idx] < self.y_max)):
            d_pa_to_split[d_idx] = 1
        else:
            d_pa_to_split[d_idx] = 0


class ParticleSplit(object):
    r"""**Hexagonal particle splitting algorithm**

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def __init__(self, pa_arr):
        r"""
        Parameters
        ----------
        pa_arr : pysph.base.particle_array.ParticleArray
            particle array of fluid

        """
        self.pa_arr = pa_arr

        # Ratio of mass of daughter particle located at center of hexagon to
        # that of its parents mass
        self.center_pa_mass_frac = 0.178705766141917

        # Ratio of mass of daughter particle located at vertex of hexagon to
        # that of its parents mass
        self.vertex_pa_mass_frac = 0.136882287617319

        # Ratio of smoothing length of daughter particle to that of its parents
        # smoothing length
        self.pa_h_ratio = 0.9

        # Ratio of distance between center daughter particle and vertex
        # daughter particle to that of its parents smoothing length
        self.center_and_vertex_pa_separation_frac = 0.4

        # Get index of the parent particles to split
        self.idx_pa_to_split = self._get_idx_of_particles_to_split()

        # Number of daughter particles located at the vertices of hexagon after
        # splitting
        self.num_vertex_pa_after_single_split = 6

    def do_particle_split(self, solver=None):
        if not self.idx_pa_to_split.size:
            # If no particles to split then return
            return
        else:
            # Properties of parent particles to split
            h_parent = self.pa_arr.h[self.idx_pa_to_split]
            h0_parent = self.pa_arr.h0[self.idx_pa_to_split]
            m_parent = self.pa_arr.m[self.idx_pa_to_split]
            x_parent = self.pa_arr.x[self.idx_pa_to_split]
            y_parent = self.pa_arr.y[self.idx_pa_to_split]
            u_parent = self.pa_arr.u[self.idx_pa_to_split]
            v_parent = self.pa_arr.v[self.idx_pa_to_split]
            u_prev_step_parent = self.pa_arr.u_prev_step[self.idx_pa_to_split]
            v_prev_step_parent = self.pa_arr.v_prev_step[self.idx_pa_to_split]
            rho_parent = self.pa_arr.rho[self.idx_pa_to_split]
            rho0_parent = self.pa_arr.rho0[self.idx_pa_to_split]
            alpha_parent = self.pa_arr.alpha[self.idx_pa_to_split]

            # Vertex daughter particle properties update
            n = self.num_vertex_pa_after_single_split
            h_vertex_pa = self.pa_h_ratio * np.repeat(h_parent, n)
            h0_vertex_pa = self.pa_h_ratio * np.repeat(h0_parent, n)
            u_prev_step_vertex_pa = np.repeat(u_prev_step_parent, n)
            v_prev_step_vertex_pa = np.repeat(v_prev_step_parent, n)
            m_vertex_pa = self.vertex_pa_mass_frac * np.repeat(m_parent, n)
            vertex_pa_pos = self._get_vertex_pa_positions(h_parent, u_parent,
                                                          v_parent)
            x_vertex_pa = vertex_pa_pos[0] + np.repeat(x_parent, n)
            y_vertex_pa = vertex_pa_pos[1] + np.repeat(y_parent, n)

            rho0_vertex_pa = np.repeat(rho0_parent, n)
            rho_vertex_pa = np.repeat(rho_parent, n)
            alpha_vertex_pa = np.repeat(alpha_parent, n)
            parent_idx_vertex_pa = np.repeat(self.idx_pa_to_split, n)

            # Note:
            # The center daughter particle properties are set at index of
            # parent. The properties of parent needed for further calculations
            # are not changed for now

            # Center daughter particle properties update
            for idx in self.idx_pa_to_split:
                self.pa_arr.m[idx] *= self.center_pa_mass_frac
                self.pa_arr.h[idx] *= self.pa_h_ratio
                self.pa_arr.h0[idx] *= self.pa_h_ratio
                self.pa_arr.parent_idx[idx] = int(idx)

            # Update particle array to include vertex daughter particles
            self._add_vertex_pa_prop(
                h0_vertex_pa, h_vertex_pa, m_vertex_pa, x_vertex_pa,
                y_vertex_pa, rho0_vertex_pa, rho_vertex_pa,
                u_prev_step_vertex_pa, v_prev_step_vertex_pa, alpha_vertex_pa,
                parent_idx_vertex_pa)

    def _get_idx_of_particles_to_split(self):
        idx_pa_to_split = []
        for idx, val in enumerate(self.pa_arr.pa_to_split):
            if val:
                idx_pa_to_split.append(idx)
        return np.array(idx_pa_to_split)

    def _get_vertex_pa_positions(self, h_parent, u_parent, v_parent):
        # Number of particles to split
        num_of_pa_to_split = len(self.idx_pa_to_split)

        n = self.num_vertex_pa_after_single_split
        theta_vertex_pa = zeros(n)
        r = self.center_and_vertex_pa_separation_frac

        for i, theta in enumerate(range(0, 360, 60)):
            theta_vertex_pa[i] = (pi/180)*theta

        # Angle of velocity vector with horizontal
        angle_vel = np.where(
            (np.abs(u_parent) > 1e-3) | (np.abs(v_parent) > 1e-3),
            np.arctan2(v_parent, u_parent), 0
            )

        # Rotates the hexagon such that its horizontal axis aligns with the
        # direction of velocity vector
        angle_actual = (np.tile(theta_vertex_pa, num_of_pa_to_split)
                        + np.repeat(angle_vel, n))

        x = r * cos(angle_actual) * np.repeat(h_parent, n)
        y = r * sin(angle_actual) * np.repeat(h_parent, n)
        return x.copy(), y.copy()

    def _add_vertex_pa_prop(self, h0_vertex_pa, h_vertex_pa, m_vertex_pa,
                            x_vertex_pa, y_vertex_pa, rho0_vertex_pa,
                            rho_vertex_pa, u_prev_step_vertex_pa,
                            v_prev_step_vertex_pa, alpha_vertex_pa,
                            parent_idx_vertex_pa):
        vertex_pa_props = {
            'm': m_vertex_pa,
            'h': h_vertex_pa,
            'h0': h0_vertex_pa,
            'x': x_vertex_pa,
            'y': y_vertex_pa,
            'u_prev_step': u_prev_step_vertex_pa,
            'v_prev_step': v_prev_step_vertex_pa,
            'rho0': rho0_vertex_pa,
            'rho': rho_vertex_pa,
            'alpha': alpha_vertex_pa,
            'parent_idx': parent_idx_vertex_pa.astype(int)
            }

        # Add vertex daughter particles to particle array
        self.pa_arr.add_particles(**vertex_pa_props)


class DaughterVelocityEval(Equation):
    r"""**Evaluation of the daughter particle velocity after splitting
    procedure**

    .. math::

        \boldsymbol{v_k} = c_v\frac{d_N}{d_k}\boldsymbol{v_N}

    where,

    .. math::

        c_v = \dfrac{A_N}{\sum_{k=1}^{M}A_k}

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def __init__(self, dest, sources, rhow=1000.0):
        r""""
        Parameters
        ----------
        rhow : float
            constant 3-D density of water (kg/m3)

        Notes
        -----
        This equation should be called before the equation SWEOS, as the parent
        particle area is required for calculating velocities. On calling the
        SWEOS equation, the parent properties are changed to the center
        daughter particle properties.

        """
        self.rhow = rhow
        super(DaughterVelocityEval, self).__init__(dest, sources)

    def initialize(self, d_sum_Ak, d_idx, d_m, d_rho, d_u, d_v, d_uh,
                   d_vh, d_u_parent, d_v_parent, d_uh_parent, d_vh_parent,
                   d_parent_idx):
        # Stores sum of areas of daughter particles
        d_sum_Ak[d_idx] = 0.0
        d_u_parent[d_idx] = d_u[d_parent_idx[d_idx]]
        d_uh_parent[d_idx] = d_uh[d_parent_idx[d_idx]]
        d_v_parent[d_idx] = d_v[d_parent_idx[d_idx]]
        d_vh_parent[d_idx] = d_vh[d_parent_idx[d_idx]]

    def loop_all(self, d_sum_Ak, d_pa_to_split, d_parent_idx, d_idx, s_m,
                 s_rho, s_parent_idx, NBRS, N_NBRS):
        i = declare('int')
        s_idx = declare('long')
        if d_pa_to_split[d_idx]:
            for i in range(N_NBRS):
                s_idx = NBRS[i]
                if s_parent_idx[s_idx] == d_parent_idx[d_idx]:
                    # Sums the area of daughter particles who have same parent
                    # idx
                    d_sum_Ak[d_idx] += s_m[s_idx] / s_rho[s_idx]

    def post_loop(self, d_idx, d_parent_idx, d_A, d_sum_Ak, d_dw, d_rho, d_u,
                  d_uh, d_vh, d_v, d_u_parent, d_v_parent, d_uh_parent,
                  d_vh_parent, t):
        # True only for daughter particles
        if d_parent_idx[d_idx]:

            # Ratio of parent area (before split) to sum of areas of its
            # daughters (after split)
            cv = d_A[d_parent_idx[d_idx]] / d_sum_Ak[d_parent_idx[d_idx]]

            # The denominator (d_rho[d_idx]/self.rhow) represents the water
            # depth of daughter particle. d_dw[d_idx] cannot be used as
            # equation of state is called after this equation (Refer Notes in
            # the constructor)
            dw_ratio = d_dw[d_parent_idx[d_idx]] / (d_rho[d_idx]/self.rhow)

            d_u[d_idx] = cv * dw_ratio * d_u_parent[d_idx]
            d_uh[d_idx] = cv * dw_ratio * d_uh_parent[d_idx]
            d_v[d_idx] = cv * dw_ratio * d_v_parent[d_idx]
            d_vh[d_idx] = cv * dw_ratio * d_vh_parent[d_idx]
            d_parent_idx[d_idx] = 0


class FindMergeable(Equation):
    r"""**Particle merging algorithm**

    Particles are tagged for merging if the following condition is
    satisfied:

    .. math::

        (A_i < A_min) and (x_min < x_i < x_max) and (y_min < y_i < y_max)

    References
    ----------
    .. [Vacondio2013] R. Vacondio et al., "Shallow water SPH for flooding with
    dynamic particle coalescing and splitting", Advances in Water Resources,
    58 (2013), pp. 10-23

    """
    def __init__(self, dest, sources, A_min, x_min=-1e9, x_max=1e9, y_min=-1e9,
                 y_max=1e9):
        r"""
        Parameters
        ----------
        A_min : float
            minimum area below which merging is activated
        x_min : float
            minimum distance along x-direction beyond which merging is
            activated
        x_max : float
            maximum distance along x-direction beyond which merging is
            deactivated
        y_min : float
            minimum distance along y-direction beyond which merging is
            activated
        y_max : float
            maximum distance along y-direction beyond which merging is
            deactivated

        Notes
        -----
        The merging algorithm merges two particles 'a' and 'b' if the following
        conditions are satisfied:

        #. Both particles have area less than A_min
        #. Both particles lies within :math:`x_min < x_i < x_max` and
        :math:`y_min < y_i < y_max`
        #. if 'a' is the closest neighbor of 'b' and vice versa

        The merging algorithm is run every timestep

        """
        self.A_min = A_min
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        super(FindMergeable, self).__init__(dest, sources)

    def loop_all(self, d_idx, d_merge, d_closest_idx, d_x, d_y, d_h, d_A,
                 d_is_merged_pa, s_x, s_y, s_A, NBRS, N_NBRS):
        # Finds the closest neighbor of a particle and stores the index of that
        # neighbor in d_closest_idx[d_idx]
        i, closest = declare('int', 2)
        s_idx = declare('unsigned int')
        d_merge[d_idx] = 0
        d_is_merged_pa[d_idx] = 0
        xi = d_x[d_idx]
        yi = d_y[d_idx]
        rmin = d_h[d_idx] * 10.0
        closest = -1
        if (d_A[d_idx] < self.A_min and ((self.x_min < d_x[d_idx] < self.x_max)
           and (self.y_min < d_y[d_idx] < self.y_max))):
            for i in range(N_NBRS):
                s_idx = NBRS[i]
                if s_idx == d_idx:
                    continue
                xij = xi - s_x[s_idx]
                yij = yi - s_y[s_idx]
                rij = sqrt(xij*xij + yij*yij)
                if rij < rmin:
                    closest = s_idx
                    rmin = rij
        d_closest_idx[d_idx] = closest

    def post_loop(self, d_idx, d_m, d_u, d_v, d_h, d_uh, d_vh, d_closest_idx,
                  d_is_merged_pa, d_merge, d_x, d_y, SPH_KERNEL):
        idx = declare('int')
        xma = declare('matrix(3)')
        xmb = declare('matrix(3)')
        idx = d_closest_idx[d_idx]
        if idx > -1:
            # If particle 'a' is closest neighbor of 'b' and vice versa
            if d_idx == d_closest_idx[idx]:
                if d_idx < idx:
                    # The newly merged particle properties are set at index of
                    # particle 'a'
                    m_merged = d_m[d_idx] + d_m[idx]
                    x_merged = ((d_m[d_idx]*d_x[d_idx] + d_m[idx]*d_x[idx])
                                / m_merged)
                    y_merged = ((d_m[d_idx]*d_y[d_idx] + d_m[idx]*d_y[idx])
                                / m_merged)
                    xma[0] = x_merged - d_x[d_idx]
                    xma[1] = y_merged - d_y[d_idx]
                    xmb[0] = x_merged - d_x[idx]
                    xmb[1] = y_merged - d_y[idx]
                    rma = sqrt(xma[0]*xma[0] + xma[1]*xma[1])
                    rmb = sqrt(xmb[0]*xmb[0] + xmb[1]*xmb[1])
                    d_u[d_idx] = ((d_m[d_idx]*d_u[d_idx] + d_m[idx]*d_u[idx])
                                  / m_merged)
                    d_uh[d_idx] = (d_m[d_idx]*d_uh[d_idx]
                                   + d_m[idx]*d_uh[idx]) / m_merged
                    d_v[d_idx] = ((d_m[d_idx]*d_v[d_idx] + d_m[idx]*d_v[idx])
                                  / m_merged)
                    d_vh[d_idx] = (d_m[d_idx]*d_vh[d_idx]
                                   + d_m[idx]*d_vh[idx]) / m_merged
                    const1 = d_m[d_idx] * SPH_KERNEL.kernel(xma, rma,
                                                            d_h[d_idx])
                    const2 = d_m[idx] * SPH_KERNEL.kernel(xmb, rmb, d_h[idx])
                    d_h[d_idx] = sqrt((7*M_PI/10.) * (m_merged/(const1+const2)))
                    d_m[d_idx] = m_merged
                    # Tags the newly formed particle after merging
                    d_is_merged_pa[d_idx] = 1
                else:
                    # Tags particle 'b' for removal after merging
                    d_merge[d_idx] = 1

    def reduce(self, dst, t, dt):
        # The indices of particle 'b' are removed from particle array after
        # merging is done
        indices = declare('object')
        indices = numpy.where(dst.merge > 0)[0]
        if len(indices) > 0:
            dst.remove_particles(indices)


class InitialDensityEvalAfterMerge(Equation):
    r"""**Initial density of the newly formed particle after merging**

    .. math ::

        \rho_M = \sum_{j}^{}m_jW_{M,j}

    References
    ----------
    .. [Vacondio2013] R. Vacondio et al., "Shallow water SPH for flooding with
    dynamic particle coalescing and splitting", Advances in Water Resources,
    58 (2013), pp. 10-23

    """
    def loop_all(self, d_rho, d_idx, d_is_merged_pa, d_x, d_y, s_h, s_m, s_x,
                 d_merge, d_closest_idx, s_y, SPH_KERNEL, NBRS, N_NBRS):
        i = declare('int')
        s_idx = declare('long')
        xij = declare('matrix(3)')
        # Evaluates the initial density of the newly formed particle after
        # merging
        if d_is_merged_pa[d_idx] == 1:
            d_rho[d_idx] = 0.0
            rij = 0.0
            rho_sum = 0.0
            for i in range(N_NBRS):
                s_idx = NBRS[i]
                xij[0] = d_x[d_idx] - s_x[s_idx]
                xij[1] = d_y[d_idx] - s_y[s_idx]
                rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1])
                rho_sum += s_m[s_idx] * SPH_KERNEL.kernel(xij, rij, s_h[s_idx])
            d_rho[d_idx] += rho_sum


class EulerStep(IntegratorStep):
    """Fast but inaccurate integrator. Use this for testing"""
    def initialize(self, d_u, d_v, d_u_prev_step, d_v_prev_step, d_idx):
        d_u_prev_step[d_idx] = d_u[d_idx]
        d_v_prev_step[d_idx] = d_v[d_idx]

    def stage1(self, d_idx, d_u, d_v, d_au, d_av, d_x, d_y, dt):
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_x[d_idx] += dt * d_u[d_idx]
        d_y[d_idx] += dt * d_v[d_idx]


class SWEStep(IntegratorStep):
    """Leap frog time integration scheme"""
    def initialize(self, t, d_u, d_v, d_uh, d_vh, d_u_prev_step, d_v_prev_step,
                   d_idx):
        # Stores the velocities at previous time step
        d_u_prev_step[d_idx] = d_u[d_idx]
        d_v_prev_step[d_idx] = d_v[d_idx]

    def stage1(self, d_uh, d_vh, d_idx, d_au, d_av, dt):
        # Velocities at half time step
        d_uh[d_idx] += dt * d_au[d_idx]
        d_vh[d_idx] += dt * d_av[d_idx]

    def stage2(self, d_u, d_v, d_uh, d_vh, d_idx, d_au, d_av, d_x, d_y, dt):
        d_x[d_idx] += dt * d_uh[d_idx]
        d_y[d_idx] += dt * d_vh[d_idx]
        d_u[d_idx] = d_uh[d_idx] + dt/2.*d_au[d_idx]
        d_v[d_idx] = d_vh[d_idx] + dt/2.*d_av[d_idx]


class SWEIntegrator(Integrator):
    """Integrator for shallow water problems"""
    def one_timestep(self, t, dt):
        self.compute_accelerations()

        self.initialize()

        # Predict
        self.stage1()

        # Call any post-stage functions.
        self.do_post_stage(0.5*dt, 1)

        # Correct
        self.stage2()

        # Call any post-stage functions.
        self.do_post_stage(dt, 2)


class GatherDensityEvalNextIteration(Equation):
    r"""**Gather formulation for evaluating the density of a particle**

    .. math::

        \rho_i = \sum_{j}{m_jW(\textbf{x}_i - \textbf{x}_j, h_i)}

    References
    ----------
    .. [Hernquist and Katz, 1988] L. Hernquist and N. Katz, "TREESPH: A
    unification of SPH with the hierarcgical tree method", The Astrophysical
    Journal Supplement Series, 70 (1989), pp 419-446.

    """
    def initialize(self, d_rho, d_idx, d_rho_prev_iter):
        # Stores density of particle i of the previous iteration
        d_rho_prev_iter[d_idx] = d_rho[d_idx]
        d_rho[d_idx] = 0.0

    def loop(self, d_rho, d_idx, s_m, s_idx, WI):
        d_rho[d_idx] += s_m[s_idx] * WI


class ScatterDensityEvalNextIteration(Equation):
    r"""**Scatter formulation for evaluating the density of a particle**

    .. math::

        \rho_i = \sum_{J}{m_JW(\textbf{x}_i - \textbf{x}_j, h_j)}

    References
    ----------
    .. [Hernquist and Katz, 1988] L. Hernquist and N. Katz, "TREESPH: A
    unification of SPH with the hierarcgical tree method", The Astrophysical
    Journal Supplement Series, 70 (1989), pp 419-446.

    """
    def initialize(self, t, d_rho, d_idx, d_rho_prev_iter):
        # Stores density of particle i of the previous iteration
        d_rho_prev_iter[d_idx] = d_rho[d_idx]
        d_rho[d_idx] = 0.0

    def loop(self, d_rho, d_idx, s_m, s_idx, WJ):
        d_rho[d_idx] += s_m[s_idx] * WJ


class NonDimensionalDensityResidual(Equation):
    r"""**Non-dimensional density residual**

    .. math::

        \psi^{k+1} = \dfrac{|\rho_i^{k+1} - \rho_i^k|}{\rho_i^k}

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def __init__(self, dest, sources=None):
        super(NonDimensionalDensityResidual, self).__init__(dest, sources)

    def post_loop(self, d_psi, d_rho, d_rho_prev_iter, d_idx):
        # Non-dimensional residual
        d_psi[d_idx] = abs(d_rho[d_idx] - d_rho_prev_iter[d_idx]) \
                       / d_rho_prev_iter[d_idx]


class CheckConvergenceDensityResidual(Equation):
    r"""The iterative process is stopped once the following condition is met

    .. math::

        \psi^{k+1} < \epsilon_{\psi}

    where,

        \epsilon_{\psi} = 1e-3


    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    Notes
    -----
    If particle splitting is activated, better to use this convergence
    criteria. It can be used even if particle splitting is not activated.

    """
    def __init__(self, dest, sources=None):
        super(CheckConvergenceDensityResidual, self).__init__(dest, sources)
        self.eqn_has_converged = 0

    def initialize(self):
        self.eqn_has_converged = 0

    def reduce(self, dst, t, dt):
        epsilon = max(dst.psi)
        if epsilon <= 1e-3:
            self.eqn_has_converged = 1

    def converged(self):
        return self.eqn_has_converged


class CorrectionFactorVariableSmoothingLength(Equation):
    r"""**Correction factor in internal force due to variable smoothing
    length**

    .. math::

        \alpha_i = -\sum_j m_j r_{ij}\frac{dW_i}{dr_{ij}}

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def initialize(self, d_idx, d_alpha):
        d_alpha[d_idx] = 0.0

    def loop(self, d_alpha, d_idx, DWIJ, XIJ, s_idx, s_m):
        d_alpha[d_idx] += -s_m[s_idx] * (DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1])


class RemoveParticlesWithZeroAlpha(Equation):
    r"""Removes particles if correction factor (alpha) in internal force due to
    variable smoothing length is zero

    """
    def __init__(self, dest):
        super(RemoveParticlesWithZeroAlpha, self).__init__(dest, None)

    def post_loop(self, d_alpha, d_pa_alpha_zero, d_idx):
        if d_alpha[d_idx] == 0:
            d_pa_alpha_zero[d_idx] = 1

    def reduce(self, dst, t, dt):
        indices = declare('object')
        indices = numpy.where(dst.pa_alpha_zero > 0)[0]
        if len(indices) > 0:
            dst.remove_particles(indices)


class SummationDensity(Equation):
    r"""**Summation Density**

    .. math::

        \rho_i = \sum_{j}{m_jW(\textbf{x}_i - \textbf{x}_j, h_i)}

    """
    def initialize(self, d_summation_rho, d_idx):
        d_summation_rho[d_idx] = 0.0

    def loop(self, d_summation_rho, d_idx, s_m, s_idx, WI):
        d_summation_rho[d_idx] += s_m[s_idx] * WI


class InitialGuessDensityVacondio(Equation):
    r"""**Initial guess density to start the iterative evaluation of density
    for time step n+1**

    .. math::

        \rho_{i(0)}^{n+1} = \rho_i^n + dt\dfrac{d\rho_i}{dt}\\
        h_{i(0)}^{n+1} = h_i^n + -\dfrac{h_i^n}{\rho_i^n}\dfrac{dt}{dm}
        \dfrac{d\rho_i}{dt}

    where,

    .. math::

        \frac{d\rho_i}{dt} = \rho_i^n\sum_j\dfrac{m_j}{\rho_j}
        (\textbf{v}_i-\textbf{v}_j).\nabla W_i

    References
    ----------
    .. [VacondioSWE-SPHysics, 2013] R. Vacondio et al., SWE-SPHysics source
    code, File: SWE_SPHYsics/SWE-SPHysics_2D_v1.0.00/source/SPHYSICS_SWE_2D/
    ac_dw_var_hj_2D.f

    Note:
    If particle splitting is activated, better to use this method. It can be
    used even if particle splitting is not activated.

    """
    def __init__(self, dest, sources, dim=2):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)

        """
        self.dim = dim
        super(InitialGuessDensityVacondio, self).__init__(dest, sources)

    def initialize(self, d_arho, d_idx):
        d_arho[d_idx] = 0

    def loop(self, d_arho, d_rho, d_idx, s_m, s_rho, s_idx, d_u_prev_step,
             d_v_prev_step, s_u_prev_step, s_v_prev_step, DWI):
        tmp1 = (d_u_prev_step[d_idx]-s_u_prev_step[s_idx]) * DWI[0]
        tmp2 = (d_v_prev_step[d_idx]-s_v_prev_step[s_idx]) * DWI[1]
        d_arho[d_idx] += d_rho[d_idx] * ((s_m[s_idx]/s_rho[s_idx])*(tmp1+tmp2))

    def post_loop(self, d_rho, d_h, dt, d_arho, d_idx):
        d_rho[d_idx] += dt * d_arho[d_idx]
        d_h[d_idx] += -(dt/self.dim)*d_h[d_idx]*(d_arho[d_idx]/d_rho[d_idx])


class InitialGuessDensity(Equation):
    r"""**Initial guess density to start the iterative evaluation of density
    for time step n+1 based on properties of time step n**

    .. math::

        \rho_{I, n+1}^{(0)} = \rho_{I,n}e^{\lambda_n}

    where,

        \lambda = \dfrac{\rho_Id_m}{\alpha_I}\sum_{J}^{}m_J
        (\textbf{v}_J - \textbf{v}_I).\nabla W_I(\textbf{x}_I
        - \textbf{x}_J, h_I)

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def __init__(self, dest, sources, dim=2):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)

        """
        self.dim = dim
        super(InitialGuessDensity, self).__init__(dest, sources)

    def initialize(self, d_exp_lambda, d_idx):
        d_exp_lambda[d_idx] = 0.0

    def loop(self, d_exp_lambda, d_u_prev_step, d_v_prev_step, d_alpha, d_idx,
             s_m, s_u_prev_step, s_v_prev_step, s_idx, DWI, dt, t):
        a1 = (d_u_prev_step[d_idx]-s_u_prev_step[s_idx]) * DWI[0]
        a2 = (d_v_prev_step[d_idx]-s_v_prev_step[s_idx]) * DWI[1]
        const = (self.dim*dt) / d_alpha[d_idx]
        d_exp_lambda[d_idx] += const * (s_m[s_idx]*(a1+a2))

    def post_loop(self, t, d_rho, d_exp_lambda, d_idx):
        d_rho[d_idx] = d_rho[d_idx] * exp(d_exp_lambda[d_idx])


class UpdateSmoothingLength(Equation):
    r"""**Update smoothing length based on density**

    .. math::

        h_I^{(k)} = h_I^{0}\biggl(\dfrac{\rho_I^0}{\rho_I^{(k)}}
        \biggl)^\frac{1}{d_m}


    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def __init__(self, dest, dim=2):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)

        """
        self.dim = dim
        super(UpdateSmoothingLength, self).__init__(dest, None)

    def post_loop(self, d_h, d_h0, d_rho0, d_rho, d_idx):
        d_h[d_idx] = d_h0[d_idx] * (d_rho0[d_idx]/d_rho[d_idx])**(1./self.dim)


class DensityResidual(Equation):
    r"""**Residual of density**

    .. math::

        R(\rho^{(k)}) = \rho_I^{(k)} - \sum_{J}^{}m_J
        W_I(\textbf{x}_I - \textbf{x}_J, h_I^{(k)})

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def __init__(self, dest, sources=None):
        super(DensityResidual, self).__init__(dest, sources)

    def post_loop(self, d_rho, d_idx, d_rho_residual, d_summation_rho, t):
        d_rho_residual[d_idx] = d_rho[d_idx] - d_summation_rho[d_idx]


class DensityNewtonRaphsonIteration(Equation):
    r"""**Newton-Raphson approximate solution for the density equation at
    iteration k+1**

    .. math::

        \rho^{(k+1)} = \rho_I^{(k)}\biggl[1 - \dfrac{R_I^{(k)}d_m}{(
        R_I^{(k)} d_m + \alpha_I^k)}\biggr]

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def __init__(self, dest, sources=None, dim=2):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)

        """
        self.dim = dim
        super(DensityNewtonRaphsonIteration, self).__init__(dest, sources)

    def initialize(self, d_rho, d_rho_prev_iter, d_idx):
        d_rho_prev_iter[d_idx] = d_rho[d_idx]

    def post_loop(self, d_rho, d_idx, d_alpha, d_rho_residual):
        a1 = d_rho_residual[d_idx] * self.dim
        a2 = a1 + d_alpha[d_idx]
        const = 1 - (a1/a2)
        d_rho[d_idx] = d_rho[d_idx] * const


class CheckConvergence(Equation):
    r"""Stops the Newton-Raphson iterative procedure if the following
    convergence criteria is satisfied:

    .. math::

        \dfrac{|R_I^{(k+1)}|}{\rho_I^{(k)}} \leq \epsilon

    where,

        \epsilon = 1e-15

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    Notes
    -----
    Use this convergence criteria when using the Newton-Raphson iterative
    procedure.

    """
    def __init__(self, dest, sources=None):
        super(CheckConvergence, self).__init__(dest, sources)
        self.eqn_has_converged = 0

    def initialize(self):
        self.eqn_has_converged = 0

    def post_loop(self, d_positive_rho_residual, d_rho_residual,
                  d_rho_prev_iter,  d_idx, t):
        d_positive_rho_residual[d_idx] = abs(d_rho_residual[d_idx])

    def reduce(self, dst, t, dt):
        max_epsilon = max(dst.positive_rho_residual / dst.rho_prev_iter)
        if max_epsilon <= 1e-15:
            self.eqn_has_converged = 1

    def converged(self):
        return self.eqn_has_converged


class SWEOS(Equation):
    r"""**Update fluid properties based on density**

    References
    ----------
    .. [Rodriguez and Bonet, 2005] M. Rodriguez and J. Bonet, "A corrected
    smooth particle hydrodynamics formulation of the shallow-water equations",
    Computers and Structures, 83 (2005), pp. 1396-1410

    """
    def __init__(self, dest, sources=None, g=9.81, rhow=1000.0):
        r"""
        Parameters
        ----------
        g : float
            acceleration due to gravity
        rhow : float
            constant 3-D density of water

        """
        self.rhow = rhow
        self.g = g
        self.fac = 0.5 * (g/rhow)
        super(SWEOS, self).__init__(dest, sources)

    def post_loop(self, d_rho, d_cs, d_u, d_v, d_idx, d_p, d_dw, d_dt_cfl,
                  d_m, d_A, d_alpha):
        # Pressure
        d_p[d_idx] = self.fac * (d_rho[d_idx])**2

        # Wave speed
        d_cs[d_idx] = sqrt(self.g * d_rho[d_idx]/self.rhow)

        # Area
        d_A[d_idx] = d_m[d_idx] / d_rho[d_idx]

        # Depth of water
        d_dw[d_idx] = d_rho[d_idx] / self.rhow

        # dt = CFL * (h_min / max(dt_cfl))
        d_dt_cfl[d_idx] = d_cs[d_idx] + (d_u[d_idx]**2 + d_v[d_idx]**2)**0.5


def mu_calc(hi=1.0, hj=1.0, velij_dot_rij=1.0, rij2=1.0):
    r"""Term present in the artificial viscosity formulation (Monaghan)

    .. math::

      \mu_{ij} = \dfrac{\bar h_{ij}\textbf{v}_{ij}.\textbf{x}_{ij}}
      {|\textbf{x}_{ij}|^2 + \zeta^2}

    References
    ----------
    .. [Monaghan2005] J. Monaghan, "Smoothed particle hydrodynamics",
        Reports on Progress in Physics, 68 (2005), pp. 1703-1759.

    """
    h_bar = (hi+hj) / 2.0
    eta2 = 0.01 * hi**2
    muij = (h_bar*velij_dot_rij) / (rij2+eta2)
    return muij


def artificial_visc(alpha=1.0, rij2=1.0,  hi=1.0, hj=1.0, rhoi=1.0, rhoj=1.0,
                    csi=1.0, csj=1.0, muij=1.0):
    r"""**Artificial viscosity based stabilization term (Monaghan)**

    Activated when :math:`\textbf{v}_{ij}.\textbf{x}_{ij} < 0`

    Given by

    .. math::

    \Pi_{ij} = \dfrac{-a\bar c_{ij}\mu_{ij}+b\bar c_{ij}\mu_{ij}^2}{\rho_{ij}}

    References
    ----------
    .. [Monaghan2005] J. Monaghan, "Smoothed particle hydrodynamics",
        Reports on Progress in Physics, 68 (2005), pp. 1703-1759.

    """
    cs_bar = (csi+csj) / 2.0
    rho_bar = (rhoi+rhoj) / 2.0
    pi_visc = -(alpha*cs_bar*muij) / rho_bar
    return pi_visc


def viscosity_LF(alpha=1.0, rij2=1.0, hi=1.0, hj=1.0, rhoi=1.0, rhoj=1.0,
                 csi=1.0, csj=1.0, muij=1.0):
    r"""**Lax-Friedrichs flux based stabilization term (Ata and Soulaimani)**

    .. math::

    \Pi_{ij} = \dfrac{\bar c_{ij}\textbf{v}_{ij}.\textbf{x}_{ij}}
    {\bar\rho_{ij}\sqrt{|x_{ij}|^2 + \zeta^2}}

    References
    ----------
    .. [Ata and Soulaimani, 2004] R. Ata and A. Soulaimani, "A stabilized SPH
    method for inviscid shallow water", Int. J. Numer. Meth. Fluids, 47 (2005),
    pp. 139-159.

    Notes
    -----
    The advantage of this term is that it automatically sets the required level
    of numerical viscosity based on the Lax-Friedrichs flux. This is the
    default stabilization method.

    """
    cs_bar = (csi+csj) / 2.0
    rho_bar = (rhoi+rhoj) / 2.0
    eta2 = 0.01 * hi**2
    h_bar = (hi+hj) / 2.0
    tmp = (muij*(rij2+eta2)**0.5) / h_bar
    pi_visc = -(cs_bar*tmp) / rho_bar
    return pi_visc


class ParticleAcceleration(Equation):
    r"""**Acceleration of a particle**

    .. math::

        \textbf{a}_i = -\frac{g+\textbf{v}_i.\textbf{k}_i\textbf{v}_i
                       -\textbf{t}_i.\nabla H_i}{1+\nabla H_i.\nabla H_i}
                       \nabla H_i - \textbf{t}_i - \textbf{S}_{fi}

    where,

    .. math::

        \textbf{t}_i &= \sum_j m_j\ \biggl[\biggl(\frac{p_j}{
        \alpha_j \rho_j^2}+0.5\Pi_{ij}\biggr)\nabla W_j(\textbf{x}_i, h_j) -
        \biggl(\frac{p_i}{\alpha_i \rho_i^2}+0.5\Pi_{ij}\biggr)\nabla
        W_i(\textbf{x}_j, h_i)\biggr]

    .. math::

        \textbf{S}_f = \textbf{v}\dfrac{gn^2|\textbf{v}|}{d^{\frac{4}{3}}}

    with,

    .. math::

        \alpha_i = -\sum_j m_j r_{ij}\frac{dW_i}{dr_{ij}}

    .. math::

        n_i  = \sum_jn_j^b\overline W_i(x_i - x_j^b, h^b)V_j

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    Notes
    -----
    The acceleration term given in [Vacondio2010] has incorrect sign.

    """
    def __init__(self, dest, sources, dim=2, u_only=False, v_only=False,
                 alpha=0.0, visc_option=2, rhow=1000.0):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        u_only : bool
            motion of fluid column in x-direction only evaluated
            (Default: False)
        v_only : bool
            motion of fluid column in y-direction only evaluated
            (Default: False)
        alpha : float
            coefficient to control amount of artificial viscosity (Monaghan)
            (Default: 0.0)
        visc_option : int
            artifical viscosity (1) or Lax-Friedrichs flux (2) based
            stabilization term (Default: 2)
        rhow : float
            constant 3-D density of water

        """
        self.g = 9.81
        self.rhow = rhow
        self.ct = self.g / (2*self.rhow)
        self.dim = dim
        self.u_only = u_only
        self.v_only = v_only
        self.alpha = alpha
        if visc_option == 1:
            self.viscous_func = artificial_visc
        else:
            self.viscous_func = viscosity_LF
        super(ParticleAcceleration, self).__init__(dest, sources)

    def initialize(self, d_idx, d_tu, d_tv):
        d_tu[d_idx] = 0.0
        d_tv[d_idx] = 0.0

    def loop(self, d_x, d_y, s_x, s_y, d_rho, d_idx, s_m, s_idx, s_rho, d_m,
             DWI, DWJ, d_au, d_av, s_alpha, d_alpha, s_p, d_p, d_tu, s_dw,
             d_dw, t, s_is_wall_boun_pa, s_tu, d_tv, s_tv, d_h, s_h, d_u, s_u,
             d_v, s_v, d_cs, s_cs):
        # True if neighbor is wall boundary particle
        if s_is_wall_boun_pa[s_idx] == 1:

            # Setting artificial viscosity to zero when a particle interacts
            # with wall boundary particles
            pi_visc = 0.0

            # Setting water depth of wall boundary particles same as particle
            # interacting with it (For sufficient pressure to prevent wall
            # penetration)
            s_dw[s_idx] = d_dw[d_idx]
        else:
            uij = d_u[d_idx] - s_u[s_idx]
            vij = d_v[d_idx] - s_v[s_idx]
            xij = d_x[d_idx] - s_x[s_idx]
            yij = d_y[d_idx] - s_y[s_idx]
            rij2 = xij**2 + yij**2
            uij_dot_xij = uij * xij
            vij_dot_yij = vij * yij
            velij_dot_rij = uij_dot_xij + vij_dot_yij

            muij = mu_calc(d_h[d_idx], s_h[s_idx], velij_dot_rij, rij2)

            if velij_dot_rij < 0:
                # Stabilization term
                pi_visc = self.viscous_func(self.alpha, rij2, d_h[d_idx],
                                            s_h[s_idx], d_rho[d_idx],
                                            s_rho[s_idx], d_cs[d_idx],
                                            s_cs[s_idx], muij)
            else:
                pi_visc = 0

        tmp1 = (s_dw[s_idx]*self.rhow*self.dim) / s_alpha[s_idx]
        tmp2 = (d_dw[d_idx]*self.rhow*self.dim) / d_alpha[d_idx]

        # Internal force per unit mass
        d_tu[d_idx] += s_m[s_idx] * ((self.ct*tmp1 + 0.5*pi_visc)*DWJ[0] +
                                     (self.ct*tmp2 + 0.5*pi_visc)*DWI[0])

        d_tv[d_idx] += s_m[s_idx] * ((self.ct*tmp1 + 0.5*pi_visc)*DWJ[1] +
                                     (self.ct*tmp2 + 0.5*pi_visc)*DWI[1])

    def _get_helpers_(self):
        return [mu_calc, artificial_visc, viscosity_LF]

    def post_loop(self, d_idx, d_u, d_v, d_tu, d_tv, d_au, d_av, d_Sfx, d_Sfy,
                  d_bx, d_by, d_bxx, d_bxy, d_byy):
        vikivi = d_u[d_idx]*d_u[d_idx]*d_bxx[d_idx] \
                 + 2*d_u[d_idx]*d_v[d_idx]*d_bxy[d_idx] \
                 + d_v[d_idx]*d_v[d_idx]*d_byy[d_idx]

        tidotgradbi = d_tu[d_idx]*d_bx[d_idx] + d_tv[d_idx]*d_by[d_idx]
        gradbidotgradbi = d_bx[d_idx]**2 + d_by[d_idx]**2

        temp3 = self.g + vikivi - tidotgradbi
        temp4 = 1 + gradbidotgradbi

        if not self.v_only:
            # Acceleration along x-direction
            d_au[d_idx] = -(temp3/temp4)*d_bx[d_idx] - d_tu[d_idx] \
                          - d_Sfx[d_idx]
        if not self.u_only:
            # Acceleration along y-direction
            d_av[d_idx] = -(temp3/temp4)*d_by[d_idx] - d_tv[d_idx] \
                          - d_Sfy[d_idx]


class FluidBottomElevation(Equation):
    r"""**Bottom elevation of fluid**

    .. math::

        b_i = \sum_jb_j^b\overline{W_i}(\textbf{x}_i - \textbf{x}_j^b, h^b)V_j

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def initialize(self, d_b, d_idx):
        d_b[d_idx] = 0.0

    def loop_all(self, d_shep_corr, d_x, d_y, d_idx, s_x, s_y, s_V, s_idx, s_h,
                 SPH_KERNEL, NBRS, N_NBRS):
        # Shepard filter
        i = declare('int')
        xij = declare('matrix(3)')
        rij = 0.0
        corr_sum = 0.0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = d_x[d_idx] - s_x[s_idx]
            xij[1] = d_y[d_idx] - s_y[s_idx]
            rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1])
            corr_sum += s_V[s_idx] * SPH_KERNEL.kernel(xij, rij, s_h[s_idx])
        d_shep_corr[d_idx] = corr_sum

    def loop(self, d_b, d_idx, s_b, s_idx, WJ, s_V, RIJ):
        d_b[d_idx] += s_b[s_idx] * WJ * s_V[s_idx]

    def post_loop(self, d_b, d_shep_corr, d_idx):
        if d_shep_corr[d_idx] > 1e-14:
            d_b[d_idx] /= d_shep_corr[d_idx]


class FluidBottomGradient(Equation):
    r"""**Bottom gradient of fluid**

    .. math::

        \nabla b_i &=& \sum_j\nabla b_j^b W_i(\textbf{x}_i - \textbf{x}_j^b,
        h^b)V_j

    Notes:
    It is obtained from a simple SPH interpolation from the gradient of bed
    particles

    """
    def initialize(self, d_idx, d_bx, d_by):
        d_bx[d_idx] = 0.0
        d_by[d_idx] = 0.0

    def loop(self, d_idx, d_bx, d_by, WJ, s_idx, s_bx, s_by, s_V):
        # Bottom gradient of fluid
        d_bx[d_idx] += s_bx[s_idx] * WJ * s_V[s_idx]
        d_by[d_idx] += s_by[s_idx] * WJ * s_V[s_idx]


class FluidBottomCurvature(Equation):
    r"""Bottom curvature of fluid**

    .. math::

        \nabla^2 b_i = \sum_j\nabla^2 b_j^b W_i(\textbf{x}_i - \textbf{x}_j^b,
        h^b)V_j

    Notes:
    It is obtained from a simple SPH interpolation from the curvature of bed
    particles

    """
    def initialize(self, d_idx, d_bx, d_by, d_bxx, d_bxy, d_byy):
        d_bxx[d_idx] = 0.0
        d_bxy[d_idx] = 0.0
        d_byy[d_idx] = 0.0

    def loop(self, d_idx, d_bxx, d_bxy, d_byy, WJ, s_idx, s_bxx, s_bxy, s_byy,
             s_V):
        # Bottom curvature of fluid
        d_bxx[d_idx] += s_bxx[s_idx] * WJ * s_V[s_idx]
        d_bxy[d_idx] += s_bxy[s_idx] * WJ * s_V[s_idx]
        d_byy[d_idx] += s_byy[s_idx] * WJ * s_V[s_idx]


class BedGradient(Equation):
    r"""**Gradient of bed**

    .. math::

        \nabla b_i = \sum_jb_j^b\tilde{\nabla}W_i(\textbf{x}_i -
        \textbf{x}_j^b, h^b)V_j

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def initialize(self, d_bx, d_by, d_idx):
        d_bx[d_idx] = 0.0
        d_by[d_idx] = 0.0

    def loop(self, d_bx, d_by, d_idx, s_b, s_idx, DWJ, s_V, RIJ):
        if RIJ > 1e-6:
            # Gradient of bed
            d_bx[d_idx] += s_b[s_idx] * DWJ[0] * s_V[s_idx]
            d_by[d_idx] += s_b[s_idx] * DWJ[1] * s_V[s_idx]


class BedCurvature(Equation):
    r"""**Curvature of bed**

    .. math::

        \biggl(\dfrac{\partial^2b}{\partial x^\alpha \partial x^\beta}
        \biggr)_i = \sum_{j}^{}\biggl(4\dfrac{x_{ij}^\alphax_{ij}^\beta}
        {r_{ij}^2}-\delta^{\alpha\beta}\biggr)\dfrac{b_i - b_j^b}{
        \textbf{r}_{ij}\textbf{r}_{ij} + \eta^2}\textbf{r}_{ij}.\tilde{\nabla}
        W_i(\textbf{x}_i - \textbf{x}_j^b, h^b)V_j

    References
    ----------
    .. [Vacondio2010] R. Vacondio, B.D. Rodgers and P.K. Stansby, "Accurate
    particle splitting for smoothed particle hydrodynamics in shallow water
    with shock capturing", Int. J. Numer. Meth. Fluids, 69 (2012), pp.
    1377-1410

    """
    def initialize(self, d_bxx, d_bxy, d_byy, d_idx):
        d_bxx[d_idx] = 0.0
        d_bxy[d_idx] = 0.0
        d_byy[d_idx] = 0.0

    def loop(self, d_bxx, d_bxy, d_byy, d_b, d_idx, s_h, s_b, s_idx, XIJ, RIJ,
             DWJ, s_V):
        if RIJ > 1e-6:
            eta = 0.01 * s_h[s_idx]
            temp1 = (d_b[d_idx]-s_b[s_idx]) / (RIJ**2+eta**2)
            temp2 = XIJ[0]*DWJ[0] + XIJ[1]*DWJ[1]
            temp_bxx = ((4*XIJ[0]**2/RIJ**2)-1) * temp1
            temp_bxy = (4*XIJ[0]*XIJ[1]/RIJ**2) * temp1
            temp_byy = ((4*XIJ[1]**2/RIJ**2)-1) * temp1
            # Curvature of bed
            d_bxx[d_idx] += temp_bxx * temp2 * s_V[s_idx]
            d_bxy[d_idx] += temp_bxy * temp2 * s_V[s_idx]
            d_byy[d_idx] += temp_byy * temp2 * s_V[s_idx]


class BedFrictionSourceEval(Equation):
    r"""**Friction source term**

    .. math::

        \textbf{S}_f = \textbf{v}\dfrac{gn^2|\textbf{v}|}{d^{\frac{4}{3}}}

    where,

    .. math::

        n_i  = \sum_jn_j^b\overline W_i(x_i - x_j^b, h^b)V_j

    """
    def __init__(self, dest, sources):
        self.g = 9.8
        super(BedFrictionSourceEval, self).__init__(dest, sources)

    def initialize(self, d_n, d_idx):
        d_n[d_idx] = 0.0

    def loop(self, d_n, d_idx, s_n, s_idx, WJ, s_V, RIJ):
        if RIJ > 1e-6:
            # Manning coefficient
            d_n[d_idx] += s_n[s_idx] * WJ * s_V[s_idx]

    def post_loop(self, d_idx, d_Sfx, d_Sfy, d_u, d_v, d_n, d_dw):
        vmag = sqrt(d_u[d_idx]**2 + d_v[d_idx]**2)
        temp = (self.g*d_n[d_idx]**2*vmag) / d_dw[d_idx]**(4.0/3.0)
        # Friction source term
        d_Sfx[d_idx] = d_u[d_idx] * temp
        d_Sfy[d_idx] = d_v[d_idx] * temp


class BoundaryInnerReimannStateEval(Equation):
    r"""Evaluates the inner Riemann state of velocity and depth

    .. math::

        \textbf{v}_i^o = \sum_j\dfrac{m_j^f}{\rho_j^f}\textbf{v}_j^f\bar
        W_i(\textbf{x}_i^o - \textbf{x}_j^f, h_o)\\

        {d}_i^o = \sum_j\dfrac{m_j^f}{\rho_j^f}d_j^f\bar W_i(\textbf{x}_i^o -
        \textbf{x}_j^f, h_o)

    References
    ----------
    .. [Vacondio2012] R. Vacondio et al., "SPH modeling of shallow flow with
    open boundaries for practical flood simulation", J. Hydraul. Eng., 2012,
    138(6), pp. 530-541.

    """
    def initialize(self, d_u_inner_reimann, d_v_inner_reimann,
                   d_dw_inner_reimann, d_idx):
        d_u_inner_reimann[d_idx] = 0.0
        d_v_inner_reimann[d_idx] = 0.0
        d_dw_inner_reimann[d_idx] = 0.0

    def loop_all(self, d_shep_corr, d_x, d_y, d_idx, s_x, s_y, s_m, s_rho,
                 s_idx, d_h, SPH_KERNEL, NBRS, N_NBRS):
        # Shepard filter
        i = declare('int')
        xij = declare('matrix(3)')
        rij = 0.0
        corr_sum = 0.0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = d_x[d_idx] - s_x[s_idx]
            xij[1] = d_y[d_idx] - s_y[s_idx]
            rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1])
            corr_sum += ((s_m[s_idx]/s_rho[s_idx])
                         * SPH_KERNEL.kernel(xij, rij, d_h[d_idx]))
        d_shep_corr[d_idx] = corr_sum

    def loop(self, d_u_inner_reimann, d_v_inner_reimann, d_dw_inner_reimann,
             d_idx, WI, s_m, s_u, s_v, s_rho, s_dw, s_idx):
        tmp = WI * (s_m[s_idx]/s_rho[s_idx])
        # Riemann invariants at open boundaries
        d_u_inner_reimann[d_idx] += s_u[s_idx] * tmp
        d_v_inner_reimann[d_idx] += s_v[s_idx] * tmp
        d_dw_inner_reimann[d_idx] += s_dw[s_idx] * tmp

    def post_loop(self, d_u_inner_reimann, d_v_inner_reimann,
                  d_dw_inner_reimann, d_shep_corr, d_idx):
        if d_shep_corr[d_idx] > 1e-14:
            d_u_inner_reimann[d_idx] /= d_shep_corr[d_idx]
            d_v_inner_reimann[d_idx] /= d_shep_corr[d_idx]
            d_dw_inner_reimann[d_idx] /= d_shep_corr[d_idx]


class SubCriticalInFlow(Equation):
    r"""**Subcritical inflow condition**

    ..math ::

        d_B = \biggl[\frac{1}{2\sqrt{g}}(v_{B,n}-v_{I,n}) + \sqrt{d_I}\biggr]^2

    References
    ----------
    .. [Vacondio2012] R. Vacondio et al., "SPH modeling of shallow flow with
    open boundaries for practical flood simulation", J. Hydraul. Eng., 2012,
    138(6), pp. 530-541.

    Notes
    -----
    The velocity is imposed at the open boundary.

    """
    def __init__(self, dest, dim=2, rhow=1000.0):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        rhow : float
            constant 3-D density of water

        """
        self.g = 9.8
        self.dim = dim
        self.rhow = rhow
        super(SubCriticalInFlow, self).__init__(dest, None)

    def post_loop(self, d_dw, d_dw_inner_reimann, d_u, d_u_inner_reimann,
                  d_rho, d_alpha, d_cs, d_idx):
        const = 1. / (2.*sqrt(self.g))
        # Properties of open boundary particles
        d_dw[d_idx] = (const*(d_u_inner_reimann[d_idx] - d_u[d_idx])
                       + sqrt(d_dw_inner_reimann[d_idx]))**2
        d_rho[d_idx] = d_dw[d_idx] * self.rhow
        d_alpha[d_idx] = self.dim * d_rho[d_idx]
        d_cs[d_idx] = sqrt(self.g * d_dw[d_idx])


class SubCriticalOutFlow(Equation):
    r"""**Subcritical outflow condition**

    ..math ::

        v_{B,n} = v_{I,n} + 2\sqrt{g}(\sqrt{d_I} - \sqrt{d_B}), v_{B,t} =
        v_{I,t}

    References
    ----------
    .. [Vacondio2012] R. Vacondio et al., "SPH modeling of shallow flow with
    open boundaries for practical flood simulation", J. Hydraul. Eng., 2012,
    138(6), pp. 530-541.

    Notes:
    -----
    The constant water depth is imposed at the open boundary.

    """
    def __init__(self, dest, dim=2, rhow=1000.0):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        rhow : float
            constant 3-D density of water

        """
        self.g = 9.8
        self.dim = dim
        self.rhow = rhow
        super(SubCriticalOutFlow, self).__init__(dest, None)

    def post_loop(self, d_dw, d_dw_inner_reimann, d_u, d_u_inner_reimann,
                  d_rho, d_cs, d_alpha, d_v, d_v_inner_reimann, d_idx):
        const = 2. * sqrt(self.g)
        # Velocities of open boundary particles
        d_u[d_idx] = (d_u_inner_reimann[d_idx]
                      + const*(sqrt(d_dw_inner_reimann[d_idx])
                               - sqrt(d_dw[d_idx])))
        d_v[d_idx] = d_v_inner_reimann[d_idx]


class SubCriticalTimeVaryingOutFlow(Equation):
    r"""**Subcritical outflow condition**

    ..math ::

        v_{B,n} = v_{I,n} + 2\sqrt{g}(\sqrt{d_I} - \sqrt{d_B}), v_{B,t} =
        v_{I,t}

    References
    ----------
    .. [Vacondio2012] R. Vacondio et al., "SPH modeling of shallow flow with
    open boundaries for practical flood simulation", J. Hydraul. Eng., 2012,
    138(6), pp. 530-541.

    Notes:
    -----
    The time varying water depth is imposed at the open boundary.

    """
    def __init__(self, dest, dim=2, rhow=1000.0):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        rhow : float
            constant 3-D density of water

        """
        self.g = 9.8
        self.dim = dim
        self.rhow = rhow
        super(SubCriticalTimeVaryingOutFlow, self).__init__(dest, None)

    def post_loop(self, d_dw, d_dw_inner_reimann, d_u, d_u_inner_reimann,
                  d_rho, d_cs, d_alpha, d_v, d_v_inner_reimann, d_idx,
                  d_dw_at_t):
        # Properties of open boundary particles
        # Time varying water depth imposed
        d_dw[d_idx] = d_dw_at_t[d_idx]
        d_rho[d_idx] = d_dw[d_idx] * self.rhow
        d_cs[d_idx] = sqrt(d_dw[d_idx] * self.g)
        d_alpha[d_idx] = d_rho[d_idx] * self.dim

        const = 2. * sqrt(self.g)
        d_u[d_idx] = (d_u_inner_reimann[d_idx]
                      + const*(sqrt(d_dw_inner_reimann[d_idx])
                               - sqrt(d_dw[d_idx])))
        d_v[d_idx] = d_v_inner_reimann[d_idx]


class SuperCriticalOutFlow(Equation):
    r"""**Supercritical outflow condition**

    .. math::

        v_{B,n} = v_{I,n}, v_{B,t} = v_{I,t}, d_B = d_I

    References
    ----------
    .. [Vacondio2012] R. Vacondio et al., "SPH modeling of shallow flow with
    open boundaries for practical flood simulation", J. Hydraul. Eng., 2012,
    138(6), pp. 530-541.

    Notes:
    -----
    For supercritical outflow condition, the velocity and water depth at the
    open boundary equals the respective inner Riemann state values. For
    supercritical inflow condition, both the velocity and water depth at the
    open boundary have to be imposed.

    """
    def __init__(self, dest, dim=2, rhow=1000.0):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        rhow : float
            constant 3-D density of water

        """
        self.g = 9.8
        self.dim = dim
        self.rhow = rhow
        super(SuperCriticalOutFlow, self).__init__(dest, None)

    def post_loop(self, d_dw, d_rho, d_dw_inner_reimann, d_u_inner_reimann,
                  d_u, d_v, d_v_inner_reimann, d_alpha, d_cs, d_idx):
        # Properties of open boundary particles
        d_u[d_idx] = d_u_inner_reimann[d_idx]
        d_v[d_idx] = d_v_inner_reimann[d_idx]
        d_dw[d_idx] = d_dw_inner_reimann[d_idx]
        d_rho[d_idx] = d_dw[d_idx] * self.rhow
        d_alpha[d_idx] = self.dim * d_rho[d_idx]
        d_cs[d_idx] = sqrt(self.g * d_dw[d_idx])


class GradientCorrectionPreStep(Equation):
    def __init__(self, dest, sources, dim=2):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)

        """
        self.dim = dim
        super(GradientCorrectionPreStep, self).__init__(dest, sources)

    def initialize(self, d_idx, d_m_mat):
        i = declare('int')
        for i in range(9):
            d_m_mat[9*d_idx + i] = 0.0

    def loop_all(self, d_idx, d_m_mat, s_V, d_x, d_y, d_z, d_h, s_x,
                 s_y, s_z, s_h, SPH_KERNEL, NBRS, N_NBRS):
        x = d_x[d_idx]
        y = d_y[d_idx]
        z = d_z[d_idx]
        h = d_h[d_idx]
        i, j, s_idx, n = declare('int', 4)
        xij = declare('matrix(3)')
        dwij = declare('matrix(3)')
        n = self.dim
        for k in range(N_NBRS):
            s_idx = NBRS[k]
            xij[0] = x - s_x[s_idx]
            xij[1] = y - s_y[s_idx]
            xij[2] = z - s_z[s_idx]
            hij = (h + s_h[s_idx]) * 0.5
            r = sqrt(xij[0]*xij[0] + xij[1]*xij[1] + xij[2]*xij[2])
            SPH_KERNEL.gradient(xij, r, hij, dwij)
            dw = sqrt(dwij[0]*dwij[0] + dwij[1]*dwij[1]
                      + dwij[2]*dwij[2])
            V = s_V[s_idx]
            if r >= 1.0e-12:
                for i in range(n):
                    xi = xij[i]
                    for j in range(n):
                        xj = xij[j]
                        d_m_mat[9*d_idx + 3*i + j] += (dw*V*xi*xj) / r


class GradientCorrection(Equation):
    r"""**Kernel Gradient Correction**

    .. math::
            \nabla \tilde{W}_{ab} = L_{a}\nabla W_{ab}

    .. math::
            L_{a} = \left(\sum \frac{m_{b}}{\rho_{b}}\nabla W_{ab}
            \mathbf{\times}x_{ab} \right)^{-1}

    References
    ----------
    .. [Bonet and Lok, 1999]  J. Bonet and T.-S.L. Lok, "Variational and
    Momentum Preservation Aspects of Smoothed Particle Hydrodynamic
    Formulations", Comput. Methods Appl. Mech. Engrg., 180 (1999), pp. 97-115

    """
    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def __init__(self, dest, sources, dim=2, tol=0.5):
        r"""
        Parameters
        ----------
        dim : int
            number of space dimensions (Default: 2)
        tol : float
            tolerance for gradient correction (Default: 0.5)

        """
        self.dim = dim
        self.tol = tol
        super(GradientCorrection, self).__init__(dest, sources)

    def loop(self, d_idx, d_m_mat, DWJ, s_h, s_idx):
        i, j, n = declare('int', 3)
        n = self.dim
        temp = declare('matrix(9)')
        aug_m = declare('matrix(12)')
        res = declare('matrix(3)')
        eps = 1.0e-04 * s_h[s_idx]
        for i in range(n):
            for j in range(n):
                temp[n*i + j] = d_m_mat[9*d_idx + 3*i + j]
        augmented_matrix(temp, DWJ, n, 1, n, aug_m)
        gj_solve(aug_m, n, 1, res)
        change = 0.0
        for i in range(n):
            change += abs(DWJ[i]-res[i]) / (abs(DWJ[i])+eps)
        if change <= self.tol:
            for i in range(n):
                DWJ[i] = res[i]


class RemoveOutofDomainParticles(Equation):
    r"""Removes particles if the following condition is satisfied:

    .. math::

        (x_i < x_min) or (x_i > x_max) or (y_i < y_min) or (y_i > y_max)

    """
    def __init__(self, dest, x_min=-1e9, x_max=1e9,
                 y_min=-1e9, y_max=1e9):
        r"""
        Parameters
        ----------
        x_min : float
            minimum distance along x-direction below which particles are
            removed
        x_max : float
            maximum distance along x-direction above which particles are
            removed
        y_min : float
            minimum distance along y-direction below which particles are
            removed
        y_max : float
            maximum distance along x-direction above which particles are
            removed

        """
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        super(RemoveOutofDomainParticles, self).__init__(dest, None)

    def initialize(self, d_pa_out_of_domain, d_x, d_y, d_idx):
        if (
            (d_x[d_idx] < self.x_min or d_x[d_idx] > self.x_max)
            or (d_y[d_idx] < self.y_min or d_y[d_idx] > self.y_max)
        ):
            d_pa_out_of_domain[d_idx] = 1
        else:
            d_pa_out_of_domain[d_idx] = 0

    def reduce(self, dst, t, dt):
        indices = declare('object')
        indices = numpy.where(dst.pa_out_of_domain > 0)[0]
        # Removes the out of domain particles
        if len(indices) > 0:
            dst.remove_particles(indices)


class RemoveCloseParticlesAtOpenBoundary(Equation):
    r"""Removes the newly created open boundary particle if the distance
    between this particle and any of its neighbor is less than min_dist_ob

    The following cases creates new open boundary particles

    * Particles which are moved back to the inlet after exiting the inlet.
    * Particles which have moved from another domain into the open boundary and
    have been converted to open boundary particles.

    References
    ----------
    .. [VacondioSWE-SPHysics, 2013] R. Vacondio et al., SWE-SPHysics source
    code, File: SWE_SPHYsics/SWE-SPHysics_2D_v1.0.00/source/SPHYSICS_SWE_2D/
    check_limits_2D.f

    """
    def __init__(self, dest, sources, min_dist_ob=0.0):
        """
        Parameters
        ----------
        min_dist_ob : float
            minimum distance of a newly created open boundary particle and its
            neighbor below which the particle is removed
        """
        self.min_dist_ob = min_dist_ob
        super(RemoveCloseParticlesAtOpenBoundary, self).__init__(dest, sources)

    def loop_all(self, d_idx, d_ob_pa_to_tag, d_ob_pa_to_remove, d_x, d_y, s_x,
                 s_y, NBRS, N_NBRS):
        i = declare('int')
        s_idx = declare('unsigned int')
        # ob_pa_to_tag is 1 for newly created open boundary particles
        if d_ob_pa_to_tag[d_idx]:
            xi = d_x[d_idx]
            yi = d_y[d_idx]
            for i in range(N_NBRS):
                s_idx = NBRS[i]
                if s_idx == d_idx:
                    continue
                xij = xi - s_x[s_idx]
                yij = yi - s_y[s_idx]
                rij = sqrt(xij*xij + yij*yij)
                if rij < self.min_dist_ob:
                    d_ob_pa_to_remove[d_idx] = 1

    def reduce(self, dst, t, dt):
        indices = declare('object')
        indices = numpy.where(dst.ob_pa_to_remove > 0)[0]
        if len(indices) > 0:
            dst.remove_particles(indices)
        dst.ob_pa_to_tag = numpy.zeros_like(dst.ob_pa_to_tag)


class RemoveFluidParticlesWithNoNeighbors(Equation):
    r"""Removes fluid particles if there exists no neighboring particles within
    its kernel radius (2*smoothing length)

    """
    def loop_all(self, d_idx, d_ob_pa_to_tag, d_fluid_pa_to_remove, d_x, d_y,
                 s_x, s_y, d_h, NBRS, N_NBRS):
        i, n_nbrs_outside_ker = declare('int', 2)
        s_idx = declare('unsigned int')
        xi = d_x[d_idx]
        yi = d_y[d_idx]
        # Number of neighbors outside the particles kernel radius
        n_nbrs_outside_ker = 0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            if s_idx == d_idx:
                continue
            xij = xi - s_x[s_idx]
            yij = yi - s_y[s_idx]
            rij = sqrt(xij*xij + yij*yij)
            if rij > 2*d_h[d_idx]:
                n_nbrs_outside_ker += 1
        # If all neighbors outside its kernel then tag particle for removal
        if n_nbrs_outside_ker == N_NBRS-1:
            d_fluid_pa_to_remove[d_idx] = 1
        else:
            d_fluid_pa_to_remove[d_idx] = 0

    def reduce(self, dst, t, dt):
        indices = declare('object')
        indices = numpy.where(dst.fluid_pa_to_remove > 0)[0]
        if len(indices) > 0:
            dst.remove_particles(indices)


class SWEInletOutletStep(IntegratorStep):
    r"""Stepper for both inlet and outlet particles for the cases dealing with
    shallow water flows

    """
    def initialize(self):
        pass

    def stage1(self, d_idx, d_x, d_y, d_uh, d_vh, d_u, d_v, dt):
        dtb2 = 0.5*dt
        d_uh[d_idx] = d_u[d_idx]
        d_vh[d_idx] = d_v[d_idx]
        d_x[d_idx] += dtb2 * d_u[d_idx]
        d_y[d_idx] += dtb2 * d_v[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_u, d_v, dt):
        dtb2 = 0.5*dt
        d_x[d_idx] += dtb2 * d_u[d_idx]
        d_y[d_idx] += dtb2 * d_v[d_idx]


class SWEInlet(object):
    """This inlet is used for shallow water flows. It has particles
    stacked along a particular axis (defaults to 'x'). These particles can
    move along any direction and as they flow out of the domain they are copied
    into the destination particle array at each timestep.

    Inlet particles are stacked by subtracting the spacing amount from the
    existing inlet array. These are copied when the inlet is created. The
    particles that cross the inlet domain are copied over to the destination
    particle array and moved back to the other side of the inlet.

    The particles from the source particle array which have moved to the inlet
    domain are removed from the source and added to the inlet particle array.

    The motion of the particles can be along any direction required.  One
    can set the 'u' velocity to have a parabolic profile in the 'y' direction
    if so desired.

    """
    def __init__(self, inlet_pa, dest_pa, source_pa, spacing, n=5, axis='x',
                 xmin=-1.0, xmax=1.0, ymin=-1.0, ymax=1.0, callback=None):
        """Constructor.

        Note that the inlet must be defined such that the spacing times the
        number of stacks of particles is equal to the length of the domain in
        the stacked direction.  For example, if particles are stacked along
        the 'x' axis and n=5 with spacing 0.1, then xmax - xmin should be 0.5.

        Parameters
        ----------

        inlet_pa: ParticleArray
           Particle array for the inlet particles.

        dest_pa: ParticleArray
           Particle array for the destination into which inlet flows.

        source_pa : ParticleArray
            Particle array from which the particles flow in.

        spacing: float
           Spacing of particles in the inlet domain.

        n: int
           Total number of copies of the initial particles.

        axis: str
           Axis along which to stack particles, one of 'x', 'y'.

        xmin, xmax, ymin, ymax : float
           Domain of the outlet.

        """
        self.inlet_pa = inlet_pa
        self.dest_pa = dest_pa
        self.spacing = spacing
        self.source_pa = source_pa
        self.callback = callback
        assert axis in ('x', 'y')
        self.axis = axis
        self.n = n
        self.xmin, self.xmax = xmin, xmax
        self.ymin, self.ymax = ymin, ymax
        self._create_inlet_particles()

    def _create_inlet_particles(self):
        props = self.inlet_pa.get_property_arrays()
        inlet_props = {}
        for prop, array in props.items():
            new_array = np.array([], dtype=array.dtype)
            for i in range(1, self.n):
                if prop == self.axis:
                    new_array = np.append(new_array, array - i*self.spacing)
                else:
                    new_array = np.append(new_array, array)
            inlet_props[prop] = new_array
        self.inlet_pa.add_particles(**inlet_props)

    def update(self, t, dt, stage):
        """This is called by the solver after each timestep and is passed
        the solver instance.
        """
        pa_add = {}
        inlet_pa = self.inlet_pa
        xmin, xmax, ymin, ymax = self.xmin, self.xmax, self.ymin, self.ymax
        lx, ly = xmax - xmin, ymax - ymin
        x, y = inlet_pa.x, inlet_pa.y

        xcond = (x > xmax)
        ycond = (y > ymax)
        # All the indices of particles which have left.
        all_idx = np.where(xcond | ycond)[0]
        # The indices which need to be wrapped around.
        x_idx = np.where(xcond)[0]
        y_idx = np.where(ycond)[0]

        # Adding particles to the destination array.
        props = inlet_pa.get_property_arrays()
        for prop, array in props.items():
            pa_add[prop] = np.array(array[all_idx])
        self.dest_pa.add_particles(**pa_add)

        # Moving the moved particles back to the array beginning.
        inlet_pa.x[x_idx] -= np.sign(inlet_pa.x[x_idx] - xmax)*lx
        inlet_pa.y[y_idx] -= np.sign(inlet_pa.y[y_idx] - ymax)*ly

        # Tags the particles which have been moved back to inlet. These tagged
        # particles are then used for checking minimum spacing condition
        # with other open boundary particles.
        inlet_pa.ob_pa_to_tag[all_idx] = 1

        source_pa = self.source_pa
        x, y = source_pa.x, source_pa.y
        idx = np.where((x <= xmax) & (x >= xmin) & (y <= ymax) & (y >=
                       ymin))[0]

        # Adding particles to the destination array.
        pa_add = {}
        props = source_pa.get_property_arrays()
        for prop, array in props.items():
            pa_add[prop] = np.array(array[idx])

        # Tags the particles which have been added to the destination array
        # from the source array. These tagged particles are then used for
        # checking minimum spacing condition with other open boundary
        # particles.
        pa_add['ob_pa_to_tag'] = np.ones_like(pa_add['ob_pa_to_tag'])

        if self.callback is not None:
            self.callback(inlet_pa, pa_add)

        inlet_pa.add_particles(**pa_add)

        source_pa.remove_particles(idx)

        # Removing the particles that moved out of inlet
        x, y = inlet_pa.x, inlet_pa.y
        idx = np.where((x > xmax) | (x < xmin) | (y > ymax) | (y < ymin))[0]
        inlet_pa.remove_particles(idx)



from __future__ import division
import numpy as np
import copy
from pysph.base.nnps import LinkedListNNPS
from pysph.base.utils import get_particle_array, get_particle_array_wcsph
from cyarray.api import UIntArray
from numpy.linalg import norm, matrix_power
from pysph.sph.equation import Equation
from pysph.tools.sph_evaluator import SPHEvaluator
from pysph.base.particle_array import ParticleArray



def distance(point1, point2=np.array([0.0, 0.0, 0.0])):
    return np.sqrt(sum((point1 - point2) * (point1 - point2)))


def distance_2d(point1, point2=np.array([0.0, 0.0])):
    return np.sqrt(sum((point1 - point2) * (point1 - point2)))


def matrix_exp(matrix):
    """
    Exponential of a matrix.

    Finds the exponential of a square matrix of any order using the
    formula exp(A) = I + (A/1!) + (A**2/2!) + (A**3/3!) + .........

    Parameters
    ----------
    matrix : numpy matrix of order nxn (square) filled with numbers

    Returns
    -------
    result : numpy matrix of the same order

    Examples
    --------
    >>>A = np.matrix([[1, 2],[2, 3]])
    >>>matrix_exp(A)
    matrix([[19.68002699, 30.56514746],
            [30.56514746, 50.24517445]])
    >>>B = np.matrix([[0, 0],[0, 0]])
    >>>matrix_exp(B)
    matrix([[1., 0.],
            [0., 1.]])
    """

    matrix = np.asarray(matrix)
    tol = 1.0e-16
    result = matrix_power(matrix, 0)
    n = 1
    condition = True
    while condition:
        adding = matrix_power(matrix, n) / (1.0 * np.math.factorial(n))
        result += adding
        residue = np.sqrt(np.sum(np.square(adding)) /
                          np.sum(np.square(result)))
        condition = (residue > tol)
        n += 1
    return result


def extrude(x, y, dx=0.01, extrude_dist=1.0, z_center=0.0):
    """
    Extrudes a 2d geometry.

    Takes a 2d geometry with x, y values and extrudes it in z direction by the
    amount extrude_dist with z_center as center

    Parameters
    ----------
    x : 1d array object with numbers
    y : 1d array object with numbers
    dx : a number
    extrude_dist : a number
    z_center : a number

    x, y should be of the same length and no x, y pair should be the same

    Returns
    -------
    x_new : 1d numpy array object with new x values
    y_new : 1d numpy array object with new y values
    z_new : 1d numpy array object with z values

    x_new, y_new, z_new are of the same length

    Examples
    --------
    >>>x = np.array([0.0])
    >>>y = np.array([0.0])
    >>>extrude(x, y, 0.1, 0.2, 0.0)
    (array([ 0., 0., 0.]),
     array([ 0., 0., 0.]),
     array([-0.1, 0., 0.1]))
    """

    z = np.arange(z_center - extrude_dist / 2.,
                  z_center + (extrude_dist + dx) / 2., dx)
    x_new = np.tile(np.asarray(x), len(z))
    y_new = np.tile(np.asarray(y), len(z))
    z_new = np.repeat(z, len(x))
    return x_new, y_new, z_new


def translate(x, y, z, x_translate=0.0, y_translate=0.0, z_translate=0.0):
    """
    Translates set of points in 3d cartisean space.

    Takes set of points and translates each and every point by some
    mentioned amount in all the 3 directions.

    Parameters
    ----------
    x : 1d array object with numbers
    y : 1d array object with numbers
    z : 1d array object with numbers
    x_translate : a number
    y_translate : a number
    z_translate : a number

    Returns
    -------
    x_new : 1d numpy array object with new x values
    y_new : 1d numpy array object with new y values
    z_new : 1d numpy array object with new z values

    Examples
    --------
    >>>x = np.array([0.0, 1.0, 2.0])
    >>>y = np.array([-1.0, 0.0, 1.5])
    >>>z = np.array([0.5, -1.5, 0.0])
    >>>translate(x, y, z, 1.0, -0.5, 2.0)
    (array([ 1., 2., 3.]), array([-1.5, -0.5, 1.]), array([2.5, 0.5, 2.]))
    """

    x_new = np.asarray(x) + x_translate
    y_new = np.asarray(y) + y_translate
    z_new = np.asarray(z) + z_translate
    return x_new, y_new, z_new


def rotate(x, y, z, axis=np.array([0.0, 0.0, 1.0]), angle=90.0):
    """
    Rotates set of points in 3d cartisean space.

    Takes set of points and rotates each point with some angle w.r.t
    a mentioned axis.

    Parameters
    ----------
    x : 1d array object with numbers
    y : 1d array object with numbers
    z : 1d array object with numbers
    axis : 1d array with 3 numbers
    angle(in degrees) : number

    Returns
    -------
    x_new : 1d numpy array object with new x values
    y_new : 1d numpy array object with new y values
    z_new : 1d numpy array object with new z values

    Examples
    --------
    >>>x = np.array([0.0, 1.0, 2.0])
    >>>y = np.array([-1.0, 0.0, 1.5])
    >>>z = np.array([0.5, -1.5, 0.0])
    >>>axis = np.array([0.0, 0.0, 1.0])
    >>>rotate(x, y, z, axis, 90.0)
    (array([ 1.00000000e+00,  4.25628483e-17, -1.50000000e+00]),
     array([-4.25628483e-17,  1.00000000e+00,  2.00000000e+00]),
     array([ 0.5, -1.5,  0. ]))
    """

    theta = angle * np.pi / 180.0
    unit_vector = np.asarray(axis) / norm(np.asarray(axis))
    matrix = np.cross(np.eye(3), unit_vector * theta)
    rotation_matrix = matrix_exp(matrix)
    new_points = []
    for xi, yi, zi in zip(np.asarray(x), np.asarray(y), np.asarray(z)):
        point = np.array([xi, yi, zi])
        new = np.dot(rotation_matrix, point)
        new_points.append(new)
    new_points = np.array(new_points)
    x_new = new_points[:, 0]
    y_new = new_points[:, 1]
    z_new = new_points[:, 2]
    return x_new, y_new, z_new


def get_2d_wall(dx=0.01, center=np.array([0.0, 0.0]), length=1.0,
                num_layers=1, up=True):
    """
    Generates a 2d wall which is parallel to x-axis. The wall can be
    rotated parallel to any axis using the rotate function. 3d wall
    can be also generated using the extrude function after generating
    particles using this function.

        ^
        |
        |
       y|*******************
        |   wall particles
        |
        |____________________>
                 x

    Parameters
    ----------
    dx : a number which is the spacing required
    center : 1d array like object which is the center of wall
    length : a number which is the length of the wall
    num_layers : Number of layers for the wall
    up : True if the layers have to created on top of base wall

    Returns
    -------
    x : 1d numpy array with x coordinates of the wall
    y : 1d numpy array with y coordinates of the wall
    """

    x = np.arange(-length / 2., length / 2. + dx, dx) + center[0]
    y = np.ones_like(x) * center[1]
    value = 1 if up else -1
    for i in range(1, num_layers):
        y1 = np.ones_like(x) * center[1] + value * i * dx
        y = np.concatenate([y, y1])
    return np.tile(x, num_layers), y


def get_2d_tank(dx=0.05, base_center=np.array([0.0, 0.0]), length=1.0,
                height=1.0, num_layers=1, outside=True, staggered=False,
                top=False):
    """
    Generates an open 2d tank with the base parallel to x-axis and the side
    walls parallel to y-axis. The tank can be rotated to any direction using
    rotate function. 3d tank can be generated using extrude function.

        ^
        |*               *
        |*   2d tank     *
       y|*  particles    *
        |*               *
        |* * * * * * * * *
        |      base
        |____________________>
                 x

    Parameters
    ----------
    dx : a number which is the spacing required
    base_center : 1d array like object which is the center of base wall
    length : a number which is the length of the base
    height : a number which is the length of the side wall
    num_layers : Number of layers for the tank
    outside : A boolean value which decides if the layers are inside or outside
    staggered : A boolean value which decides if the layers are staggered or not
    top : A boolean value which decides if the top is present or not

    Returns
    -------
    x : 1d numpy array with x coordinates of the tank
    y : 1d numpy array with y coordinates of the tank
    """

    dy = dx
    fac = 1 if outside else 0
    if staggered:
        dx = dx/2

    start = fac*(1 - num_layers)*dx
    end = fac*num_layers*dx + (1 - fac) * dx
    x, y = np.mgrid[start:length+end:dx, start:height+end:dy]

    topset = 0 if top else 10*height
    if staggered:
        topset += dx
        y[1::2] += dx

    offset = 0 if outside else (num_layers-1)*dx
    cond = ~((x > offset) & (x < length-offset) &
             (y > offset) & (y < height+topset-offset))
    return x[cond] + base_center[0] - length/2, y[cond] + base_center[1]


def get_2d_circle(dx=0.01, r=0.5, center=np.array([0.0, 0.0])):
    """
    Generates a completely filled 2d circular area.

    Parameters
    ----------
    dx : a number which is the spacing required
    r : a number which is the radius of the circle
    center : 1d array like object which is the center of the circle

    Returns
    -------
    x : 1d numpy array with x coordinates of the circle particles
    y : 1d numpy array with y coordinates of the circle particles
    """

    N = int(2.0 * r / dx) + 1
    x, y = np.mgrid[-r:r:N * 1j, -r:r:N * 1j]
    x, y = np.ravel(x), np.ravel(y)
    condition = (x * x + y * y <= r * r)
    x, y = x[condition], y[condition]
    return x + center[0], y + center[1]


def get_2d_hollow_circle(dx=0.01, r=1.0, center=np.array([0.0, 0.0]),
                         num_layers=2, inside=True):
    """
    Generates a hollow 2d circle with some number of layers either on the
    inside or on the outside of the body which is taken as an argument

    Parameters
    ----------
    dx : a number which is the spacing required
    r : a number which is the radius of the circle
    center : 1d array like object which is the center of the circle
    num_layers : a number (int)
    inside : boolean (True or False). If this is True then the layers
             are generated inside the circle

    Returns
    -------
    x : 1d numpy array with x coordinates of the circle particles
    y : 1d numpy array with y coordinates of the circle particles
    """

    r_grid = r + dx * num_layers
    N = int(2.0 * r_grid / dx) + 1
    x, y = np.mgrid[-r_grid:r_grid:N * 1j, -r_grid:r_grid:N * 1j]
    x, y = np.ravel(x), np.ravel(y)
    if inside:
        cond1 = (x * x + y * y <= r * r)
        cond2 = (x * x + y * y >= (r - num_layers * dx)**2)
    else:
        cond1 = (x * x + y * y >= r * r)
        cond2 = (x * x + y * y <= (r + num_layers * dx)**2)
    cond = cond1 & cond2
    x, y = x[cond], y[cond]
    return x + center[0], y + center[0]


def get_3d_hollow_cylinder(dx=0.01, r=0.5, length=1.0,
                           center=np.array([0.0, 0.0, 0.0]),
                           num_layers=2, inside=True):
    """
    Generates a 3d hollow cylinder which is a extruded geometry
    of the hollow circle with a closed base.

    Parameters
    ----------
    dx : a number which is the spacing required
    r : a number which is the radius of the cylinder
    length : a number which is the length of the cylinder
    center : 1d array like object which is the center of the cylinder
    num_layers : a number (int)
    inside : boolean (True or False). If this is True then the layers
             are generated inside the cylinder

    Returns
    -------
    x : 1d numpy array with x coordinates of the cylinder particles
    y : 1d numpy array with y coordinates of the cylinder particles
    z : 1d numpy array with z coordinates of the cylinder particles
    """

    x_2d, y_2d = get_2d_hollow_circle(dx, r, center[:2], num_layers, inside)
    x, y, z = extrude(x_2d, y_2d, dx, length - dx, center[2] + dx / 2.)
    x_circle, y_circle = get_2d_circle(dx, r, center[:2])
    z_circle = np.ones_like(x_circle) * (center[2] - length / 2.)
    x = np.concatenate([x, x_circle])
    y = np.concatenate([y, y_circle])
    z = np.concatenate([z, z_circle])
    return x, y, z


def get_2d_block(dx=0.01, length=1.0, height=1.0, center=np.array([0., 0.])):
    """
    Generates a 2d rectangular block of particles with axes parallel to
    the coordinate axes.

     ^
     |
     |h * * * * * * *
     |e * * * * * * *
    y|i * * * * * * *
     |g * * * * * * *
     |h * * * * * * *
     |t * * * * * * *
     |  * * * * * * *
     |    length
     |________________>
                      x

    Parameters
    ----------
    dx : a number which is the spacing required
    length : a number which is the length of the block
    height : a number which is the height of the block
    center : 1d array like object which is the center of the block

    Returns
    -------
    x : 1d numpy array with x coordinates of the block particles
    y : 1d numpy array with y coordinates of the block particles
    """

    n1 = int(length / dx) + 1
    n2 = int(height / dx) + 1
    x, y = np.mgrid[-length / 2.:length / 2.:n1 *
                    1j, -height / 2.:height / 2.:n2 * 1j]
    x, y = np.ravel(x), np.ravel(y)
    return x + center[0], y + center[1]


def get_3d_sphere(dx=0.01, r=0.5, center=np.array([0.0, 0.0, 0.0])):
    """
    Generates a 3d sphere.

    Parameters
    ----------
    dx : a number which is the spacing required
    r : a number which is the radius of the sphere
    center : 1d array like object which is the center of the sphere

    Returns
    -------
    x : 1d numpy array with x coordinates of the sphere particles
    y : 1d numpy array with y coordinates of the sphere particles
    z : 1d numpy array with z coordinates of the sphere particles
    """

    N = int(2.0 * r / dx) + 1
    x, y, z = np.mgrid[-r:r:N * 1j, -r:r:N * 1j, -r:r:N * 1j]
    x, y, z = np.ravel(x), np.ravel(y), np.ravel(z)
    cond = (x * x + y * y + z * z <= r * r)
    x, y, z = x[cond], y[cond], z[cond]
    return x + center[0], y + center[1], z + center[2]


def get_3d_block(dx=0.01, length=1.0, height=1.0, depth=1.0,
                 center=np.array([0., 0., 0.])):
    """
    Generates a 3d block of particles with the length, height and depth
    parallel to x, y and z axis respectively.

    Paramters
    ---------
    dx : a number which is the spacing required
    length : a number which is the length of the block
    height : a number which is the height of the block
    depth : a number which is the depth of the block
    center : 1d array like object which is the center of the block

    Returns
    -------
    x : 1d numpy array with x coordinates of the block particles
    y : 1d numpy array with y coordinates of the block particles
    z : 1d numpy array with z coordinates of the block particles
    """

    n1 = int(length / dx) + 1
    n2 = int(height / dx) + 1
    n3 = int(depth / dx) + 1
    x, y, z = np.mgrid[-length / 2.:length / 2.:n1 * 1j, -height /
                       2.:height / 2.:n2 * 1j, -depth / 2.:depth / 2.:n3 * 1j]
    x, y, z = np.ravel(x), np.ravel(y), np.ravel(z)
    return x + center[0], y + center[1], z + center[2]


def get_4digit_naca_airfoil(dx=0.01, airfoil='0012', c=1.0):
    """
    Generates a 4 digit series NACA airfoil. For a 4 digit series airfoil,
    the first digit is the (maximum camber / chord) * 100, second digit is
    (location of maximum camber / chord) * 10 and the third and fourth digits
    are the (maximum thickness / chord) * 100. The particles generated
    using this function will form a solid 2d airfoil.

    Parameters
    ----------
    dx : a number which is the spacing required
    airfoil : a string of 4 characters which is the airfoil name
    c : a number which is the chord of the airfoil

    Returns
    -------
    x : 1d numpy array with x coordinates of the airfoil particles
    y : 1d numpy array with y coordinates of the airfoil particles

    References
    ----------
    https://en.wikipedia.org/wiki/NACA_airfoil
    """

    n = int(c / dx) + 1
    x, y = np.mgrid[0:c:n * 1j, -c / 2.:c / 2.:n * 1j]
    x = np.ravel(x)
    y = np.ravel(y)
    x_naca = []
    y_naca = []
    t = float(airfoil[2:]) * 0.01 * c
    if airfoil[:2] == '00':
        for xi, yi in zip(x, y):
            yt = 5.0 * t * (0.2969 * np.sqrt(xi / c) - 0.1260 * (xi / c) -
                            0.3516 * ((xi / c)**2.) + 0.2843 * ((xi / c)**3.)
                            - 0.1015 * ((xi / c)**4.))
            if abs(yi) <= yt:
                x_naca.append(xi)
                y_naca.append(yi)
    else:
        m = 0.01 * float(airfoil[0])
        p = 0.1 * float(airfoil[1])
        for xi, yi in zip(x, y):
            yt = 5.0 * t * (0.2969 * np.sqrt(xi / c) - 0.1260 * (xi / c) -
                            0.3516 * ((xi / c)**2.) + 0.2843 * ((xi / c)**3.)
                            - 0.1015 * ((xi / c)**4.))
            if xi <= p * c:
                yc = (m / (p * p)) * (2. * p * (xi / c) - (xi / c)**2.)
                dydx = (2. * m / (p * p)) * (p - xi / c) / c
            else:
                yc = (m / ((1. - p) * (1. - p))) * \
                    (1. - 2. * p + 2. * p * (xi / c) - (xi / c)**2.)
                dydx = (2. * m / ((1. - p) * (1. - p))) * (p - xi / c) / c
            theta = np.arctan(dydx)
            if yi >= 0.0:
                yu = yc + yt * np.cos(theta)
                if yi <= yu:
                    xu = xi - yt * np.sin(theta)
                    x_naca.append(xu)
                    y_naca.append(yi)
            else:
                yl = yc - yt * np.cos(theta)
                if yi >= yl:
                    xl = xi + yt * np.sin(theta)
                    x_naca.append(xl)
                    y_naca.append(yi)
    x_naca = np.array(x_naca)
    y_naca = np.array(y_naca)
    return x_naca, y_naca


def _get_m_k(series):
    if series == '210':
        return 0.058, 361.4
    elif series == '220':
        return 0.126, 51.64
    elif series == '230':
        return 0.2025, 15.957
    elif series == '240':
        return 0.290, 6.643
    elif series == '250':
        return 0.391, 3.23
    elif series == '221':
        return 0.130, 51.99
    elif series == '231':
        return 0.217, 15.793
    elif series == '241':
        return 0.318, 6.52
    elif series == '251':
        return 0.441, 3.191


def get_5digit_naca_airfoil(dx=0.01, airfoil='23112', c=1.0):
    """
    Generates a 5 digit series NACA airfoil. For a 5 digit series airfoil,
    the first digit is the design lift coefficient * 20 / 3, second digit is
    (location of maximum camber / chord) * 20, third digit indicates the
    reflexitivity of the camber and the fourth and fifth digits are the
    (maximum thickness / chord) * 100. The particles generated using this
    function will form a solid 2d airfoil.

    Parameters
    ----------
    dx : a number which is the spacing required
    airfoil : a string of 5 characters which is the airfoil name
    c : a number which is the chord of the airfoil

    Returns
    -------
    x : 1d numpy array with x coordinates of the airfoil particles
    y : 1d numpy array with y coordinates of the airfoil particles

    References
    ----------
    https://en.wikipedia.org/wiki/NACA_airfoil
    http://www.aerospaceweb.org/question/airfoils/q0041.shtml
    """

    n = int(c / dx) + 1
    x, y = np.mgrid[0:c:n * 1j, -c / 2.:c / 2.:n * 1j]
    x = np.ravel(x)
    y = np.ravel(y)
    x_naca = []
    y_naca = []
    t = 0.01 * float(airfoil[3:])
    series = airfoil[:3]
    m, k = _get_m_k(series)
    for xi, yi in zip(x, y):
        yt = 5.0 * t * (0.2969 * np.sqrt(xi / c) - 0.1260 * (xi / c) -
                        0.3516 * ((xi / c)**2.) + 0.2843 * ((xi / c)**3.)
                        - 0.1015 * ((xi / c)**4.))
        xn = xi / c
        if xn <= m:
            yc = c * (k / 6.) * (xn**3. - 3. * m *
                                 xn * xn + m * m * (3. - m) * xn)
            dydx = (k / 6.) * (3. * xn * xn - 6. * m * xn + m * m * (3. - m))
        else:
            yc = c * (k * (m**3.) / 6.) * (1. - xn)
            dydx = -(k * (m**3.) / 6.)
        theta = np.arctan(dydx)
        if yi >= 0.0:
            yu = yc + yt * np.cos(theta)
            if yi <= yu:
                xu = xi - yt * np.sin(theta)
                x_naca.append(xu)
                y_naca.append(yi)
        else:
            yl = yc - yt * np.cos(theta)
            if yi >= yl:
                xl = xi + yt * np.sin(theta)
                x_naca.append(xl)
                y_naca.append(yi)
    x_naca = np.array(x_naca)
    y_naca = np.array(y_naca)
    return x_naca, y_naca


def get_naca_wing(dx=0.01, airfoil='0012', span=1.0, chord=1.0):
    """
    Generates a wing using a NACA 4 or 5 digit series airfoil. This will
    generate only a rectangular wing.

    Parameters
    ----------
    dx : a number which is the spacing required
    airfoil : a string of 4 or 5 characters which is the airfoil name
    span : a number which is the span of the wing
    c : a number which is the chord of the wing

    Returns
    -------
    x : 1d numpy array with x coordinates of the airfoil particles
    y : 1d numpy array with y coordinates of the airfoil particles
    z : 1d numpy array with z coordinates of the airfoil particles
    """

    if len(airfoil) == 4:
        x, y = get_4digit_naca_airfoil(dx, airfoil, chord)
    elif len(airfoil) == 5:
        x, y = get_5digit_naca_airfoil(dx, airfoil, chord)
    return extrude(x, y, dx, span)


class FindRepeatedPoints(Equation):
    def loop_all(self, d_idx, d_min_idx, NBRS, N_NBRS):
        d_min_idx[d_idx] = NBRS[0]
        for i in range(N_NBRS):
            if NBRS[i] < d_min_idx[d_idx]:
                d_min_idx[d_idx] = NBRS[i]


def evaluate_area_of_triangle(points=[]):
    """Calculate are of a triangle

    Parameters
    ---------
    Points: array

    Returns
    -------
    Area of the triangle
    """
    from math import sqrt
    x1 = points[0, 0]
    y1 = points[0, 1]
    z1 = points[0, 2]
    x2 = points[1, 0]
    y2 = points[1, 1]
    z2 = points[1, 2]
    x3 = points[2, 0]
    y3 = points[2, 1]
    z3 = points[2, 2]
    v1 = (z2-z3) * (y2-y1) - (z2-z1) * (y2-y3)
    v2 = - (z2-z3) * (x2-x1) + (z2-z1) * (x2-x3)
    v3 = (y2-y3) * (x2-x1) - (y2-y1) * (x2-x3)
    ar = 0.5 * sqrt(v1*v1 + v2*v2 + v3*v3)
    return ar


def remove_repeated_points(x, y, z, dx_triangle):
    EPS = np.finfo(float).eps
    pa_mesh = ParticleArray(name='mesh', x=x, y=y, z=z, h=EPS)
    pa_grid = ParticleArray(name='grid', x=x, y=y, z=z, h=EPS)
    pa_mesh.add_property(name='min_idx')

    equation = [FindRepeatedPoints(dest='mesh', sources=['grid'])]
    sph_eval = SPHEvaluator([pa_mesh, pa_grid], equation, dim=3)
    sph_eval.evaluate()

    idx = list(set(pa_mesh.min_idx))
    idx = np.array(idx, dtype=int)
    return pa_mesh.x[idx], pa_mesh.y[idx], pa_mesh.z[idx]


def find_overlap_particles(fluid_parray, solid_parray, dx_solid, dim=3):
    """This function will take 2 particle arrays as input and will find all the
    particles of the first particle array which are in the vicinity of the
    particles from second particle array. The function will find all the
    particles within the dx_solid vicinity so some particles may be identified
    at the outer surface of the particles from the second particle array.

    The particle arrays should atleast contain x, y and h values for a 2d case
    and atleast x, y, z and h values for a 3d case.

    Parameters
    ----------
    fluid_parray : a pysph particle array object
    solid_parray : a pysph particle array object
    dx_solid : a number which is the dx of the second particle array
    dim : dimensionality of the problem

    Returns
    -------
    list of particle indices to remove from the first array.

    """

    x = fluid_parray.x
    x1 = solid_parray.x
    y = fluid_parray.y
    y1 = solid_parray.y
    z = fluid_parray.z
    z1 = solid_parray.z
    if dim == 2:
        z = np.zeros_like(x)
        z1 = np.zeros_like(x1)
    to_remove = []
    ll_nnps = LinkedListNNPS(dim, [fluid_parray, solid_parray])
    for i in range(len(x)):
        nbrs = UIntArray()
        ll_nnps.get_nearest_particles(1, 0, i, nbrs)
        point_i = np.array([x[i], y[i], z[i]])
        near_points = nbrs.get_npy_array()
        distances = []
        for ind in near_points:
            dest = [x1[ind], y1[ind], z1[ind]]
            distances.append(distance(point_i, dest))
        if len(distances) == 0:
            continue
        elif min(distances) < (dx_solid * (1.0 - 1.0e-07)):
            to_remove.append(i)
    return to_remove


def remove_overlap_particles(fluid_parray, solid_parray, dx_solid, dim=3):
    """
    This function will take 2 particle arrays as input and will remove all
    the particles of the first particle array which are in the vicinity of
    the particles from second particle array. The function will remove all
    the particles within the dx_solid vicinity so some particles are removed
    at the outer surface of the particles from the second particle array.

    The particle arrays should atleast contain x, y and h values for a 2d case
    and atleast x, y, z and h values for a 3d case

    Parameters
    ----------
    fluid_parray : a pysph particle array object
    solid_parray : a pysph particle array object
    dx_solid : a number which is the dx of the second particle array
    dim : dimensionality of the problem

    Returns
    -------
    None
    """

    idx = find_overlap_particles(fluid_parray, solid_parray, dx_solid, dim)
    fluid_parray.remove_particles(idx)


def show_2d(points, **kw):
    """Show two-dimensional geometry data.

    The `points` are a tuple of x, y, z values, the extra keyword arguments are
    passed along to the scatter function.

    """
    import matplotlib.pyplot as plt
    plt.scatter(points[0], points[1], **kw)
    plt.xlabel('X')
    plt.ylabel('Y')


def show_3d(points, **kw):
    """Show two-dimensional geometry data.

    The `points` are a tuple of x, y, z values, the extra keyword arguments are
    passed along to the `mlab.points3d` function.

    """
    from mayavi import mlab
    mlab.points3d(points[0], points[1], points[2], **kw)
    mlab.axes(xlabel='X', ylabel='Y', zlabel='Z')


def get_packed_periodic_packed_particles(add_opt_func, folder, dx, L, B, H=0,
                                         dim=2, dfreq=-1, pb=None, nu=None,
                                         k=None, tol=1e-2):
    """ Creates a periodic packed 2D or 3D domain. It creates particles which
    are not aligned but packed such that the number density is uniform.

    Parameters
    ----------
    add_opt_func : options function from the parent Application class
    folder : Application class output directory
    dx : float
        required particle spacing
    L : float
        length of the domain
    B : float
        Width of the domain
    H : float
        Height of the domain
    dim : int
        dimensionality of the problem
    dfreq : int
        projection frequency of particles
    pb : float
        background pressure (default: 1.0)
    nu : float
        viscosity coefficient (default: 0.3/dx)
    k : float
        coefficient of repulsion (default: 0.005*dx)
    tol : float
        tolerance value for convergence (default: 1e-2)

    Returns
    -------
    xs: float
        x coordinate of solid particles
    ys: float
        y coordinate of solid particles
    zs: float
        z coordinate of solid particles
    xf: float
        x coordinate of fluid particles
    yf: float
        y coordinate of fluid particles
    zf: float
        z coordinate of fluid particles
    """
    from pysph.tools.packer import get_packing_folders, readdata
    import os
    preprocess_folder, layer_folder, res_file = get_packing_folders(folder, dx)
    no_solid = True

    if os.path.exists(res_file):
        return readdata(res_file)
    else:
        from pysph.tools.packer import Packer
        packer = Packer(
            None, preprocess_folder, None, add_opt_func, dx, res_file, dim=dim,
            L=L, B=B, H=H, pb=pb, nu=nu, k=k, dfreq=dfreq, no_solid=no_solid,
            tol=tol
        )
        packer.run()
        packer.post_process(packer.info_filename)
        return readdata(res_file)


def get_packed_2d_particles_from_surface_coordinates(
        add_opt_func, folder, dx, x, y, pb=None, nu=None, k=None, scale=1.0,
        shift=False, dfreq=-1, invert_normal=False, hardpoints=None,
        use_prediction=False, filter_layers=False, reduce_dfreq=False,
        tol=1e-2):
    """ Creates a packed configuration of particles around the given
    coordinates of a 2D geometry.

    Parameters
    ----------
    add_opt_func : method
        options function from the parent Application class
    folder : string
        Application class output directory
    dx : float
        required particle spacing
    x : array
        x coordinates of the geometry
    y : array
        y coordinates of the geometry
    pb : float
        background pressure (default: 1.0)
    nu : float
        viscosity coefficient (default: 0.3/dx)
    k : float
        coefficient of repulsion (default: 0.005*dx)
    scale : float
        the scaling factor for the coordinates
    dfreq : int
        projection frequency of particles
    invert_normal : bool
        if True the computed normals are inverted
    hardpoints : dict
        the dictionary of hardpoints
    use_prediction : bool
        if True, points are projected quickly to reach prediction points
    filter_layers : bool
        if True, particles away from boundary are frozen
    reduce_dfreq : bool
        if True, reduce projection frequency
    tol : float
        tolerance value for convergence (default: 1e-2)

    Returns
    -------
    xs: float
        x coordinate of solid particles
    ys: float
        y coordinate of solid particles
    zs: float
        z coordinate of solid particles
    xf: float
        x coordinate of fluid particles
    yf: float
        y coordinate of fluid particles
    zf: float
        z coordinate of fluid particles
    """

    from pysph.tools.packer import get_packing_folders, readdata
    import os
    preprocess_folder, layer_folder, res_file = get_packing_folders(folder, dx)

    if os.path.exists(res_file):
        return readdata(res_file)
    else:
        from pysph.tools.packer import Packer, HexaToRectLayer
        packer = Packer(
            None, preprocess_folder, None, add_opt_func, dx, res_file, dim=2,
            x=x, y=y, pb=pb, nu=nu, k=k, dfreq=dfreq, hardpoints=hardpoints,
            use_prediction=use_prediction, filter_layers=filter_layers,
            reduce_dfreq=reduce_dfreq, tol=tol, scale=scale, shift=shift,
            invert_normal=invert_normal)
        packer.run()
        packer.post_process(packer.info_filename)

        hextorect = HexaToRectLayer(
            None, layer_folder, None, add_opt_func, dx, res_file, dim=2, x=x,
            y=y, pb=pb, nu=nu, k=k, dfreq=dfreq, hardpoints=hardpoints,
            use_prediction=use_prediction, filter_layers=filter_layers,
            reduce_dfreq=reduce_dfreq, tol=tol, scale=scale, shift=shift,
            invert_normal=invert_normal, no_solid=True)
        hextorect.run()
        hextorect.post_process(hextorect.info_filename)

        return readdata(res_file)


def get_packed_2d_particles_from_surface_file(
        add_opt_func, folder, dx, filename, pb=None, nu=None, k=None,
        scale=1.0, shift=False, dfreq=-1, invert_normal=False,
        hardpoints=None, use_prediction=False, filter_layers=False,
        reduce_dfreq=False, tol=1e-2):
    """ Creates a packed configuration of particles around the given geometry
    file containing the x, y coordinates.

    Parameters
    ----------
    add_opt_func : method
        options function from the parent Application class
    folder : string
        Application class output directory
    dx : float
        required particle spacing
    filename : string
        file containing the x, y coordinates of the geometry
    pb : float
        background pressure (default: 1.0)
    nu : float
        viscosity coefficient (default: 0.3/dx)
    k : float
        coefficient of repulsion (default: 0.005*dx)
    scale : float
        the scaling factor for the coordinates
    dfreq : int
        projection frequency of particles
    invert_normal : bool
        if True the computed normals are inverted
    hardpoints : dict
        the dictionary of hardpoints
    use_prediction : bool
        if True, points are projected quickly to reach prediction points
    filter_layers : bool
        if True, particles away from boundary are frozen
    reduce_dfreq : bool
        if True, reduce projection frequency
    tol : float
        tolerance value for convergence (default: 1e-2)

    Returns
    -------
    xs: float
        x coordinate of solid particles
    ys: float
        y coordinate of solid particles
    zs: float
        z coordinate of solid particles
    xf: float
        x coordinate of fluid particles
    yf: float
        y coordinate of fluid particles
    zf: float
        z coordinate of fluid particles
    """

    from pysph.tools.packer import get_packing_folders, readdata
    import os
    preprocess_folder, layer_folder, res_file = get_packing_folders(folder, dx)

    if os.path.exists(res_file):
        return readdata(res_file)
    else:
        from pysph.tools.packer import Packer, HexaToRectLayer
        packer = Packer(
            None, preprocess_folder, None, add_opt_func, dx, res_file, dim=2,
            filename=filename, pb=pb, nu=nu, k=k, dfreq=dfreq,
            hardpoints=hardpoints, use_prediction=use_prediction,
            filter_layers=filter_layers, reduce_dfreq=reduce_dfreq, tol=tol,
            scale=scale, shift=shift, invert_normal=invert_normal)
        packer.run()
        packer.post_process(packer.info_filename)

        hextorect = HexaToRectLayer(
            None, layer_folder, None, add_opt_func, dx, res_file, dim=2,
            filename=filename, pb=pb, nu=nu, k=k, dfreq=dfreq,
            hardpoints=hardpoints, use_prediction=use_prediction,
            filter_layers=filter_layers, reduce_dfreq=reduce_dfreq, tol=tol,
            scale=scale, shift=shift, invert_normal=invert_normal,
            no_solid=True)
        hextorect.run()
        hextorect.post_process(hextorect.info_filename)

        return readdata(res_file)


def get_packed_3d_particles_from_surface_file(
        add_opt_func, folder, dx, filename, pb=None, nu=None, k=None,
        scale=1.0, shift=False, dfreq=-1, invert_normal=False,
        hardpoints=None, use_prediction=False, filter_layers=False,
        reduce_dfreq=False, tol=1e-2):
    """ Creates a packed configuration of particles around the given STL
    file containing the x, y, z coordinates and normals.

    Parameters
    ----------
    add_opt_func : method
        options function from the parent Application class
    folder : string
        Application class output directory
    dx : float
        required particle spacing
    filename : string
        the STL filename
    pb : float
        background pressure (default: 1.0)
    nu : float
        viscosity coefficient (default: 0.3/dx)
    k : float
        coefficient of repulsion (default: 0.005*dx)
    scale : float
        the scaling factor for the coordinates
    dfreq : int
        projection frequency of particles
    invert_normal : bool
        if True the computed normals are inverted
    hardpoints : dict
        the dictionary of hardpoints
    use_prediction : bool
        if True, points are projected quickly to reach prediction points
    filter_layers : bool
        if True, particles away from boundary are frozen
    reduce_dfreq : bool
        if True, reduce projection frequency
    tol : float
        tolerance value for convergence (default: 1e-2)

    Returns
    -------
    xs: float
        x coordinate of solid particles
    ys: float
        y coordinate of solid particles
    zs: float
        z coordinate of solid particles
    xf: float
        x coordinate of fluid particles
    yf: float
        y coordinate of fluid particles
    zf: float
        z coordinate of fluid particles
    """
    from pysph.tools.packer import get_packing_folders, readdata
    import os
    preprocess_folder, layer_folder, res_file = get_packing_folders(folder, dx)

    if os.path.exists(res_file):
        return readdata(res_file)
    else:
        from pysph.tools.packer import Packer, HexaToRectLayer
        packer = Packer(
            None, preprocess_folder, None, add_opt_func, dx, res_file, dim=3,
            filename=filename, pb=pb, nu=nu, k=k, dfreq=dfreq,
            hardpoints=hardpoints, use_prediction=use_prediction,
            filter_layers=filter_layers, reduce_dfreq=reduce_dfreq, tol=tol,
            scale=scale, shift=shift, invert_normal=invert_normal)
        packer.run()
        packer.post_process(packer.info_filename)

        hextorect = HexaToRectLayer(
            None, layer_folder, None, add_opt_func, dx, res_file, dim=3,
            filename=filename, pb=pb, nu=nu, k=k, dfreq=dfreq,
            hardpoints=hardpoints, use_prediction=use_prediction,
            filter_layers=filter_layers, reduce_dfreq=reduce_dfreq, tol=tol,
            scale=scale, shift=shift, invert_normal=invert_normal,
            no_solid=True)
        hextorect.run()
        hextorect.post_process(hextorect.info_filename)

        return readdata(res_file)


def create_fluid_around_packing(dx, xf, yf, L, B, zf=[0.0], H=0.0, **props):
    """ Create the outer fluid particles around the generated packing. It adds
    the packed fluid particles and generate a concatenated particle array

    Parameters
    ----------
    dx : float
        particle spacing
    xf : array
        x coordinate of fluid particles
    yf : array
        y coordinate of fluid particles
    L : float
        length of the domain
    B : float
        width of the domain
    zf : array
        z coordinate of fluid particles
    H : float
        height of the domain

    Returns
    -------
    Particle array of fluid
    """
    from pysph.base.utils import get_particle_array
    xmax = max(xf)
    xmin = min(xf)
    ymax = max(yf)
    ymin = min(yf)
    zmax = max(zf)
    zmin = min(zf)

    print(props)
    if H < 1e-14:
        eps = dx/10
        x, y = np.mgrid[dx/2:L:dx, -B/2+dx/2:B/2:dx]
        cond = ~(
            (x - xmin + eps > 1e-14) &
            (x - xmax - eps < 1e-14) &
            (y - ymin + eps > 1e-14) &
            (y - ymax - eps < 1e-14)
        )
        x = np.concatenate((x[cond], xf))
        y = np.concatenate((y[cond], yf))
        return get_particle_array(
            name='fluid', x=x, y=y, **props)
    else:
        eps = dx/10
        x, y, z = np.mgrid[dx/2:L:dx, -B/2+dx/2:B/2:dx, -H/2+dx/2:H/2:dx]
        cond = ~(
            (x - xmin + eps > 1e-14) &
            (x - xmax - eps < 1e-14) &
            (y - ymin + eps > 1e-14) &
            (y - ymax - eps < 1e-14) &
            (z - zmin + eps > 1e-14) &
            (z - zmax - eps < 1e-14)
        )
        x = np.concatenate((x[cond], xf))
        y = np.concatenate((y[cond], yf))
        z = np.concatenate((z[cond], zf))
        return get_particle_array(name='fluid', x=x, y=y, z=z, **props)



"""Viscosity functions"""

from pysph.sph.equation import Equation


class LaminarViscosity(Equation):
    def __init__(self, dest, sources, nu, eta=0.01):
        self.nu = nu
        self.eta = eta
        super(LaminarViscosity, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho,
             d_au, d_av, d_aw, DWIJ, XIJ, VIJ, R2IJ, HIJ):
        rhoa = d_rho[d_idx]
        rhob = s_rho[s_idx]

        # scalar part of the kernel gradient
        Fij = DWIJ[0] * XIJ[0] + DWIJ[1] * XIJ[1] + DWIJ[2] * XIJ[2]

        mb = s_m[s_idx]

        tmp = mb * 4 * self.nu * Fij/((rhoa + rhob)*(R2IJ + self.eta*HIJ*HIJ))

        # accelerations
        d_au[d_idx] += tmp * VIJ[0]
        d_av[d_idx] += tmp * VIJ[1]
        d_aw[d_idx] += tmp * VIJ[2]


class MonaghanSignalViscosityFluids(Equation):
    def __init__(self, dest, sources, alpha, h):
        self.alpha = 0.125 * alpha * h
        super(MonaghanSignalViscosityFluids, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, d_rho, s_rho, s_m,
             d_au, d_av, d_aw, d_cs, s_cs,
             RIJ, HIJ, VIJ, XIJ, DWIJ):

        nua = self.alpha * d_cs[d_idx]
        nub = self.alpha * s_cs[s_idx]

        rhoa = d_rho[d_idx]
        rhob = s_rho[s_idx]

        mb = s_m[s_idx]

        vabdotrab = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]

        eta = nua * nub/(nua*rhoa + nub*rhob)
        force = -16 * eta * vabdotrab/(HIJ * (RIJ + 0.01*HIJ*HIJ))

        d_au[d_idx] += -mb * force * DWIJ[0]
        d_av[d_idx] += -mb * force * DWIJ[1]
        d_aw[d_idx] += -mb * force * DWIJ[2]


class ClearyArtificialViscosity(Equation):
    r"""Artificial viscosity proposed By P. Cleary:

    .. math::

       \mathcal{Pi}_{ab} = -\frac{16}{\mu_a \mu_b}{\rho_a \rho_b
       (\mu_a + \mu_b)}\left( \frac{\boldsymbol{v}_{ab} \cdot
       \boldsymbol{r}_{ab}}{\boldsymbol{r}_{ab}^2 + \epsilon} \right),

    where the viscosity is determined from the parameter
    :math:`\alpha` as

    .. math::

        \mu_a = \frac{1}{8}\alpha h_a c_a \rho_a

    This equation is described in the 2005 review paper by Monaghan

    - J. J. Monaghan, "Smoothed Particle Hydrodynamics", Reports on
      Progress in Physics, 2005, 68, pp 1703--1759 [JM05]

    """

    def __init__(self, dest, sources, dim, alpha=1.0):
        self.alpha = alpha
        self.factor = 16.0
        if dim == 3:
            self.factor = 20.0

        # Base class initialization
        super(ClearyArtificialViscosity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, s_m, d_rho, s_rho, d_h, s_h, d_cs, s_cs,
             d_au, d_av, d_aw, XIJ, VIJ, R2IJ, EPS, DWIJ):

        # viscosity parameters for each particle Eq. (8.8) in [JM05]
        mua = 0.125 * self.alpha * d_h[d_idx] * d_cs[d_idx] * d_rho[d_idx]
        mub = 0.125 * self.alpha * s_h[s_idx] * s_cs[s_idx] * s_rho[s_idx]

        # \boldsymbol{v}_{ab} \cdot \boldsymbol{r}_{ab}
        dot = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]

        # Pi_ab term. Eq. (8.9) in [JM05]
        rhoa = d_rho[d_idx]
        rhob = s_rho[s_idx]
        eta = mua*mub/(rhoa*rhob*(mua + mub))
        piab = -s_m[s_idx] * self.factor*eta * (dot/(R2IJ + EPS))

        # accelerations due to viscosity Eq. (8.2) in [JM05]
        d_au[d_idx] += piab * DWIJ[0]
        d_av[d_idx] += piab * DWIJ[1]
        d_aw[d_idx] += piab * DWIJ[2]


class LaminarViscosityDeltaSPH(Equation):
    r"""See section 2 of the below reference [Sun2017]

    References
    ----------

    .. [Sun2017] P. Sun, A. Colagrossi, S. Marrone, A. Zhang "The $\delta$
       plus-SPH model: simple procedures for a further improvement of the SPH
       scheme", Computer Methods in Applied Mechanics and Engineering 315
       (2017), pp. 25-49.

    """
    def __init__(self, dest, sources, dim, rho0, nu):
        self.dim = dim
        self.rho0 = rho0
        self.nu = nu
        super(LaminarViscosityDeltaSPH, self).__init__(dest, sources)

    def loop(self, d_idx, s_idx, s_m, s_rho, d_rho, d_au, d_av, d_aw, HIJ,
             DWIJ, R2IJ, EPS, VIJ, XIJ):
        Vj = s_m[s_idx]/s_rho[s_idx]
        rhoi = d_rho[d_idx]
        vdotxij = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]
        piij = vdotxij/(R2IJ + EPS)

        fac = 2 * (self.dim + 2) * self.nu * self.rho0 * piij * Vj / rhoi
        d_au[d_idx] += fac*DWIJ[0]
        d_av[d_idx] += fac*DWIJ[1]
        d_aw[d_idx] += fac*DWIJ[2]



""" An implementation of a general solver base class """
from __future__ import print_function
# System library imports.
import os
import numpy

from compyle.profile import profile, profile_ctx
# PySPH imports
from pysph.base.kernels import CubicSpline
from pysph.sph.acceleration_eval import make_acceleration_evals
from pysph.sph.sph_compiler import SPHCompiler

from pysph.solver.utils import ProgressBar, load, dump

import logging
logger = logging.getLogger(__name__)

EPSILON = numpy.finfo(float).eps*2


class Solver(object):
    """Base class for all PySPH Solvers
    """
    def __init__(self, dim=2, integrator=None, kernel=None,
                 n_damp=0, tf=1.0, dt=1e-3,
                 adaptive_timestep=False, cfl=0.3,
                 output_at_times=(),
                 fixed_h=False, **kwargs):
        """**Constructor**

        Any additional keyword args are used to set the values of any
        of the attributes.

        Parameters
        ----------

        dim : int
            Dimension of the problem

        integrator : pysph.sph.integrator.Integrator
            Integrator to use

        kernel : pysph.base.kernels.Kernel
            SPH kernel to use

        n_damp : int
            Number of timesteps for which the initial damping is required.
            This is used to improve stability for problems with strong
            discontinuity in initial condition.
            Setting it to zero will disable damping of the timesteps.

        dt : double
            Suggested initial time step for integration

        tf : double
            Final time for integration

        adaptive_timestep : bint
            Flag to use adaptive time steps

        cfl : double
            CFL number for adaptive time stepping

        pfreq : int
            Output files dumping frequency.

        output_at_times : list/array
            Optional list of output times to force dump the output file

        fixed_h : bint
            Flag for constant smoothing lengths `h`

        reorder_freq : int
            The number of iterations after which particles should
            be re-ordered.  If zero, do not do this.

        Example
        -------

        >>> integrator = PECIntegrator(fluid=WCSPHStep())
        >>> kernel = CubicSpline(dim=2)
        >>> solver = Solver(dim=2, integrator=integrator, kernel=kernel,
        ...                 n_damp=50, tf=1.0, dt=1e-3, adaptive_timestep=True,
        ...                 pfreq=100, cfl=0.5, output_at_times=[1e-1, 1.0])

        """

        self.integrator = integrator
        self.dim = dim
        if kernel is not None:
            self.kernel = kernel
        else:
            self.kernel = CubicSpline(dim)

        # set the particles to None
        self.particles = None

        self.acceleration_evals = None
        self.nnps = None

        # solver time and iteration count
        self.t = 0
        self.count = 0

        self.execute_commands = None

        # list of functions to be called before and after an integration step
        self.pre_step_callbacks = []
        self.post_step_callbacks = []

        # List of functions to be called after each stage of the integrator.
        self.post_stage_callbacks = []

        # default output printing frequency
        self.pfreq = 100

        # Compress generated files.
        self.compress_output = False
        self.disable_output = False

        # the process id for parallel runs
        self.pid = None

        # set the default rank to 0
        self.rank = 0

        # set the default comm to None.
        self.comm = None

        # set the default mode to serial
        self.in_parallel = False

        # arrays to print output
        self.arrays_to_print = []

        # the default parallel output mode
        self.parallel_output_mode = "collected"

        # flag to print all arrays
        self.detailed_output = False

        # flag to save Remote arrays
        self.output_only_real = True

        # output filename
        self.fname = self.__class__.__name__

        # output drectory
        self.output_directory = self.fname+'_output'

        # solution damping to avoid impulsive starts
        self.n_damp = n_damp

        # Use adaptive time steps and cfl number
        self.adaptive_timestep = adaptive_timestep
        self.cfl = cfl

        # list of output times
        self.output_at_times = numpy.asarray(output_at_times)
        self.force_output = False

        # default time step constants
        self.tf = tf
        self.dt = dt
        self.max_steps = 1 << 31
        self._prev_dt = None
        self._damping_factor = 1.0
        self._epsilon = EPSILON*tf

        # flag for constant smoothing lengths
        self.fixed_h = fixed_h

        self.reorder_freq = 0

        # Set all extra keyword arguments
        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)
            else:
                msg = 'Unknown keyword arg "%s" passed to constructor' % attr
                raise TypeError(msg)

    ##########################################################################
    # Public interface.
    ##########################################################################
    def setup(self, particles, equations, nnps, kernel=None, fixed_h=False):
        """ Setup the solver.

        The solver's processor id is set if the in_parallel flag is set
        to true.

        The order of the integrating calcs is determined by the solver's
        order attribute.

        This is usually called at the start of a PySPH simulation.

        """

        self.particles = particles
        if kernel is not None:
            self.kernel = kernel

        mode = 'mpi' if self.in_parallel else 'serial'
        self.acceleration_evals = make_acceleration_evals(
            particles, equations, self.kernel, mode
        )

        sph_compiler = SPHCompiler(
            self.acceleration_evals, self.integrator
        )
        sph_compiler.compile()

        # Set the nnps for all concerned objects.
        self.nnps = nnps
        for ae in self.acceleration_evals:
            ae.set_nnps(nnps)
        self.integrator.set_nnps(nnps)

        # set the parallel manager for the integrator
        self.integrator.set_parallel_manager(self.pm)

        # Set the post_stage_callback.
        self.integrator.set_post_stage_callback(self._post_stage_callback)

        # set integrator option for constant smoothing length
        self.fixed_h = fixed_h
        self.integrator.set_fixed_h(fixed_h)

        logger.debug("Solver setup complete.")

    def add_post_stage_callback(self, callback):
        """These callbacks are called *after* each integrator stage.

        The callbacks are passed (current_time, dt, stage).  See the the
        `Integrator.one_timestep` methods for examples of how this is called.

        Example
        -------

        >>> def post_stage_callback_function(t, dt, stage):
        >>>     # This function is called after every stage of integrator.
        >>>     print(t, dt, stage)
        >>>     # Do something
        >>> solver.add_post_stage_callback(post_stage_callback_function)
        """
        self.post_stage_callbacks.append(callback)

    def add_post_step_callback(self, callback):
        """These callbacks are called *after* each timestep is performed.

        The callbacks are passed the solver instance (i.e. self).

        Example
        -------

        >>> def post_step_callback_function(solver):
        >>>     # This function is called after every time step.
        >>>     print(solver.t, solver.dt)
        >>>     # Do something
        >>> solver.add_post_step_callback(post_step_callback_function)
        """
        self.post_step_callbacks.append(callback)

    def add_pre_step_callback(self, callback):
        """These callbacks are called *before* each timestep is performed.

        The callbacks are passed the solver instance (i.e. self).

        Example
        -------

        >>> def pre_step_callback_function(solver):
        >>>     # This function is called before every time step.
        >>>     print(solver.t, solver.dt)
        >>>     # Do something
        >>> solver.add_pre_step_callback(pre_step_callback_function)
        """
        self.pre_step_callbacks.append(callback)

    def append_particle_arrrays(self, arrays):
        """ Append the particle arrays to the existing particle arrays
        """
        if not self.particles:
            print('Warning! Particles not defined.')
            return

        for array in self.particles:
            array_name = array.name
            for arr in arrays:
                if array_name == arr.name:
                    array.append_parray(arr)

        self.setup(self.particles)

    @profile
    def reorder_particles(self):
        """Re-order particles so as to coalesce memory access.
        """
        for i in range(len(self.particles)):
            self.nnps.spatially_order_particles(i)
        # We must update after the reorder.
        self.nnps.update()

    def set_adaptive_timestep(self, value):
        """Set it to True to use adaptive timestepping based on
        cfl, viscous and force factor.

        Look at pysph.sph.integrator.compute_time_step for more details.
        """
        self.adaptive_timestep = value

    def set_cfl(self, value):
        'Set the CFL number for adaptive time stepping'
        self.cfl = value

    def set_final_time(self, tf):
        """ Set the final time for the simulation """
        self.tf = tf
        self._epsilon = EPSILON*tf

    def set_n_damp(self, ndamp):
        """Set the number of timesteps for which the timestep should be
        initially damped.
        """
        self.n_damp = ndamp

    def set_time_step(self, dt):
        """ Set the time step to use """
        self.dt = dt

    def set_print_freq(self, n):
        """ Set the output print frequency """
        self.pfreq = n

    def set_disable_output(self, value):
        """Disable file output.
        """
        self.disable_output = value

    def set_arrays_to_print(self, array_names=None):
        """Only print the arrays with the given names.
        """

        available_arrays = [array.name for array in self.particles]

        if array_names:
            for name in array_names:
                if name not in available_arrays:
                    raise RuntimeError("Array %s not availabe" % (name))

                for arr in self.particles:
                    if arr.name == name:
                        array = arr
                        break
                self.arrays_to_print.append(array)
        else:
            self.arrays_to_print = self.particles

    def set_output_fname(self, fname):
        """ Set the output file name """
        self.fname = fname

    def set_output_printing_level(self, detailed_output):
        """ Set the output printing level """
        self.detailed_output = detailed_output

    def set_output_only_real(self, output_only_real):
        """ Set the flag to save out only real particles """
        self.output_only_real = output_only_real

    def set_output_directory(self, path):
        """ Set the output directory """
        self.output_directory = path

    def set_output_at_times(self, output_at_times):
        """ Set a list of output times """
        self.output_at_times = numpy.asarray(output_at_times)

    def set_max_steps(self, max_steps):
        """Set the maximum number of iterations to perform.
        """
        self.max_steps = max_steps

    def set_compress_output(self, compress):
        """Compress the dumped output files.
        """
        self.compress_output = compress

    def set_parallel_output_mode(self, mode="collected"):
        """Set the default solver dump mode in parallel.

        The available modes are:

        collected : Collect array data from all processors on root and
                    dump a single file.


        distributed : Each processor dumps a file locally.

        """
        assert mode in ("collected", "distributed")
        self.parallel_output_mode = mode

    def set_command_handler(self, callable, command_interval=1):
        """ set the `callable` to be called at every `command_interval`
        iteration

        the `callable` is called with the solver instance as an argument
        """
        self.execute_commands = callable
        self.command_interval = command_interval

    def set_parallel_manager(self, pm):
        self.pm = pm

    def set_reorder_freq(self, freq):
        """Set the reorder frequency in number of iterations.
        """
        self.reorder_freq = freq

    def barrier(self):
        if self.comm:
            self.comm.barrier()

    def solve(self, show_progress=True):
        """ Solve the system

        Notes
        -----
        Pre-stepping functions are those that need to be called before
        the integrator is called.

        Similarly, post step functions are those that are called after
        the stepping within the integrator.

        """
        if self.in_parallel:
            show = False
        else:
            show = show_progress
        bar = ProgressBar(self.t, self.tf, show=show)
        self._epsilon = EPSILON*self.tf

        # Initial solution
        self.dump_output()
        self.barrier()  # everybody waits for this to complete

        reorder_freq = self.reorder_freq
        if reorder_freq > 0:
            self.reorder_particles()

        # Compute the accelerations once for the predictor corrector
        # integrator to work correctly at the first time step.
        self.integrator.initial_acceleration(self.t, self.dt)

        # Now get a suitable adaptive (if requested) and damped timestep to
        # integrate with.
        self.dt = self._get_timestep()

        while (self.tf - self.t) > self._epsilon and \
              (self.count < self.max_steps):

            # perform any pre step functions
            if self.pre_step_callbacks:
                with profile_ctx('Solver.pre_step_callback'):
                    for callback in self.pre_step_callbacks:
                        callback(self)

            if self.rank == 0:
                logger.debug(
                    "Iteration=%d, time=%f, timestep=%f" %
                    (self.count, self.t, self.dt)
                )
            # perform the integration and update the time.
            # print('Solver Iteration', self.count, self.dt, self.t)
            self.integrator.step(self.t, self.dt)

            # perform any post step functions
            if self.post_step_callbacks:
                with profile_ctx('Solver.post_step_callback'):
                    for callback in self.post_step_callbacks:
                        callback(self)

            # update time and iteration counters if successfully
            # integrated
            self.t += self.dt
            self.count += 1
            self._epsilon = EPSILON*self.tf*self.count

            # Compute the next timestep.
            self.dt = self._get_timestep()

            # Note: this may adjust dt to land at a desired time.
            self._dump_output_if_needed()

            # update progress bar
            bar.update(self.t)

            # update the time for all arrays
            self.update_particle_time()

            if reorder_freq > 0 and (self.count % reorder_freq == 0):
                self.reorder_particles()

            if self.execute_commands is not None:
                if self.count % self.command_interval == 0:
                    self.execute_commands(self)

        # close the progress bar
        bar.finish()

        # final output save
        self.dump_output()

    def update_particle_time(self):
        for array in self.particles:
            array.set_time(self.t)

    @profile
    def dump_output(self):
        """Dump the simulation results to file

        The arrays used for printing are determined by the particle
        array's `output_property_arrays` data attribute. For debugging
        it is sometimes nice to have all the arrays (including
        accelerations) saved. This can be chosen from using the
        command line option `--detailed-output`

        Output data Format:

        A single file named as: <fname>_<rank>_<iteration_count>.npz

        The data is saved as a Python dictionary with two keys:

        `solver_data` : Solver meta data like time, dt and iteration number

        `arrays` : A dictionary keyed on particle array names and with
                   particle properties as value.

         Example:

         You can load the data output by PySPH like so:

         >>> from pysph.solver.utils import load
         >>> data = load('output_directory/filename_x_xxx.npz')
         >>> solver_data = data['solver_data']
         >>> arrays = data['arrays']
         >>> fluid = arrays['fluid']
         >>> ...

         In the above example, it is assumed that the output file
         contained an array named fluid.

        """
        if self.disable_output:
            return

        if self.rank == 0:
            msg = 'Writing output at time %g, iteration %d, dt = %g' % (
                self.t, self.count, self.dt)
            logger.info(msg)

        fname = os.path.join(self.output_directory,
                             '%s_%05d' % (self.fname, self.count))

        comm = None
        if self.parallel_output_mode == "collected" and self.in_parallel:
            comm = self.comm

        dump(fname, self.particles, self._get_solver_data(),
             detailed_output=self.detailed_output,
             only_real=self.output_only_real, mpi_comm=comm,
             compress=self.compress_output)

    def load_output(self, count):
        """Load particle data from dumped output file.

        Parameters
        ----------
        count : str
            The iteration time from which to load the data. If time is '?' then
            list of available data files is returned else the latest available
            data file is used

        Notes
        -----

        Data is loaded from the :py:attr:`output_directory` using the same
        format as stored by the :py:meth:`dump_output` method. Proper
        functioning required that all the relevant properties of arrays be
        dumped.

        """
        # get the list of available files
        available_files = [i.rsplit('_', 1)[1][:-4]
                           for i in os.listdir(self.output_directory)
                           if i.startswith(self.fname) and i.endswith('.npz')]

        if count == '?':
            return sorted(set(available_files), key=int)

        else:
            if count not in available_files:
                msg = "File with iteration count `%s` does not exist" % (count)
                msg += "\nValid iteration counts are %s" % (
                    sorted(set(available_files), key=int))
                raise IOError(msg)

        array_names = [pa.name for pa in self.particles]

        # load the output file
        data = load(os.path.join(self.output_directory,
                                 self.fname+'_'+str(count)+'.npz'))

        arrays = [data["arrays"][i] for i in array_names]

        # set the Particle's arrays
        self.particles = arrays

        solver_data = data['solver_data']

        self.t = float(solver_data['t'])
        self.dt = float(solver_data['dt'])
        self.count = int(solver_data['count'])

    def get_options(self, arg_parser):
        """ Implement this to add additional options for the application """
        pass

    def setup_solver(self, options=None):
        """ Implement the basic solvers here

        All subclasses of Solver may implement this function to add the
        necessary operations for the problem at hand.

        Parameters
        ----------
        options : dict
            options set by the user using commandline (there is no guarantee
            of existence of any key)
        """
        pass

    ##########################################################################
    # Non-public interface.
    ##########################################################################
    def _compute_timestep(self):
        undamped_dt = self._get_undamped_timestep()
        if self.adaptive_timestep:
            # locally stable time step
            dt = self.integrator.compute_time_step(undamped_dt, self.cfl)

            # set the globally stable time step across all processors
            if self.in_parallel:
                if dt is None:
                    # For some reason this processor does not have an adaptive
                    # timestep constraint so we set it to a large number so the
                    # timestep is determined by the other processors.
                    dt = 1e20
                dt = self.pm.update_time_steps(dt)
            else:
                if dt is None:
                    dt = undamped_dt
        else:
            dt = undamped_dt

        return dt

    def _damp_timestep(self, dt):
        """Damp the timestep initially to prevent transient errors at startup.

        This basically damps the initial timesteps by the factor

        0.5  (sin(pi*(-0.5 + count/n_damp)) + 1)

        Where n_damp is the number of iterations to damp the timestep for and
        count is the number of iterations.

        """
        n_damp = self.n_damp
        if self.count < n_damp and n_damp > 0:
            iter_fraction = (self.count+1)/float(n_damp)
            fac = 0.5*(numpy.sin(numpy.pi*(-0.5 + iter_fraction)) + 1.0)
            self._damping_factor = fac
        else:
            self._damping_factor = 1.0

        return dt*self._damping_factor

    def _dump_output_if_needed(self):
        """Dump output if needed while solve is running.

        This is called by `solve`.

        Warning
        -------

        This will adjust `dt` if the user has asked for output at a
        non-integral multiple of dt.
        """
        if abs(self.t - self.tf) < self._epsilon:
            return

        # dump output if the iteration number is a multiple of the printing
        # frequency.
        dump = self.count % self.pfreq == 0

        # Consider the other cases if user has requested output at a specified
        # time.

        output_at_times = self.output_at_times
        dt = self.dt

        # adjust dt to land on specific output times or dump output if we have
        # reached a desired time.
        if len(output_at_times) > 0:
            tdiff = output_at_times - self.t

            if numpy.any(numpy.abs(tdiff) < self._epsilon):
                dump = True

            # Our next step may exceed a required timestep so we adjust the
            # timestep.
            timestep_too_big = (tdiff > 0.0) & (tdiff < dt)
            if numpy.any(timestep_too_big):
                indices = numpy.where(timestep_too_big)[0]
                index = indices[0]
                output_time = output_at_times[index]
                if ((abs(output_time - self.t) < self._epsilon) and
                   (len(indices) > 1)):
                    index = indices[1]
                    output_time = output_at_times[index]
                if abs(output_time - self.t) > self._epsilon:
                    # It sometimes happens that the current time is just
                    # shy of the requested output time which results in a
                    # ridiculously small dt so we skip that case.

                    # Compute the new time-step to fall on the specified output
                    # time instant and save the previous dt value.
                    self._prev_dt = dt
                    self.dt = float(output_time - self.t)

        if dump:
            self.dump_output()
            self.barrier()

    def _get_solver_data(self):
        if self._prev_dt is not None:
            dt = self._prev_dt/self._damping_factor
        else:
            dt = self._get_undamped_timestep()

        return {'dt': dt, 't': self.t, 'count': self.count}

    @profile
    def _get_timestep(self):
        if abs(self.tf - self.t) < self._epsilon:
            # We have reached the end, so no need to adjust the timestep
            # anymore.
            return self.dt

        if self._prev_dt is not None and \
           abs(self._prev_dt - self.dt) > self._epsilon:
            # if the _prev_dt was set then we need to use it as the current dt
            # was set to print at an intermediate time.
            self.dt = self._prev_dt
            self._prev_dt = None

        dt = self._compute_timestep()
        dt = self._damp_timestep(dt)

        # adjust dt to land exactly on final time
        if (self.t + dt) > (self.tf - self._epsilon):
            dt = self.tf - self.t

        return dt

    def _get_undamped_timestep(self):
        return self.dt/self._damping_factor

    def _post_stage_callback(self, time, dt, stage):
        if self.post_stage_callbacks:
            with profile_ctx('Solver.post_stage_callback'):
                for callback in self.post_stage_callbacks:
                    callback(time, dt, stage)


############################################################################
